"""
This module provides functions for analyzing tempo characteristics in musical performances.
It includes calculations for tempo slope, tempo drift, and tempo fluctuation.

<small>**Author**: Huw Cheston (2025)</small>

References
----------

- Cheston, H., Schlichting, J. L., Cross, I., & Harrison, P. M. C. (2024).
      Jazz Trio Database: Automated Annotation of Jazz Piano Trio Recordings 
      Processed Using Audio Source Separation. *Transactions of the International
      Society for Music Information Retrieval*, 7(1), 144–158.
      https://doi.org/10.5334/tismir.186

- Cheston, H., Cross, I., & Harrison, P. (2024). Trade-offs in Coordination Strategies
      for Duet Jazz Performances Subject to Network Delay and Jitter.
      *Music Perception*, 42(1), 48–72. https://doi.org/10.1525/mp.2024.42.1.48
"""

from typing import Iterable, Union

import numpy as np

__author__ = "Huw Cheston"


# flake8: noqa: W605
def tempo_slope(beats: Iterable[float]) -> float:
    r"""
    Calculates the tempo slope for a sequence of beat timestamps.

    The tempo slope represents the overall tempo change per second in a performance.
    It is determined by the slope of a linear regression of instantaneous tempo against
    beat onset time. A negative slope indicates deceleration, while a positive slope
    indicates acceleration. The units are (quarter-note) beats-per-minute-per-second

    The equation is:

    $\hat{S} = \frac{\sum\limits_{i=1}^N (x_i - \bar{x}) (y_i - \bar{y})}{\sum\limits_{i=1}^N (x_i - \bar{x})^2}$,

    where $x_i$ is the time of beat $i$ and $y_i$ is the tempo value in
    (quarter-note) beats-per-minute.

    Parameters
    ----------
    beats : Iterable[float]
        An iterable of beat timestamps in seconds, such as quarter-note onsets.

    Returns
    -------
    float
        The computed tempo slope value.

    """

    return float(fit_tempo_linear_regression(beats).slope)


# flake8: noqa: W605
def tempo_fluctuation(beats: Iterable[float]) -> float:
    r"""
    Calculates the fluctuation around the overall tempo of a sequence of beats.

    Tempo fluctuation is measured as the standard deviation of the instantaneous tempo,
    normalized by the mean tempo. Higher values indicate greater variability in tempo.

    The equation is:

    $\text{F} = \dfrac{\sqrt{\frac{1}{N-1}
     \sum\limits_{i=1}^N (y_i - \bar{y})^2}}{\bar{y}}$,

    where $y_i$ is the tempo value in (quarter-note) beats-per-minute at beat $i$.

    Parameters
    ----------
    beats : Iterable[float]
        An iterable of beat timestamps in seconds, such as quarter-note onsets.

    Returns
    -------
    float
        The computed tempo fluctuation value.

    """

    # Compute instantaneous tempo values: this will also validate the input
    tempos = beats_to_tempo(beats)
    # Compute tempo fluctuation
    # No need for np.nanstd etc. here, we won't have NaN values
    return float(np.std(tempos) / np.mean(tempos))


def tempo_mean(beats: Iterable[float]):
    r"""
    Calculates the mean tempo from an iterable of timestamps in quarter-note beats-per-minute.

    The mean tempo can be calculated simply as:

    $\bar{y} = \dfrac{\sum\limits_{i=1}^N\frac{60}{x_i - x_{i-1}}}{N-1}$

    where $x_i$ is the time of beat $i$ (and $i \geq 1$) and $N$ is the number of beats.

    Parameters
    ----------
    beats : Iterable[float]
        An iterable of beat timestamps in seconds, such as quarter-note onsets.

    Returns
    -------
    float
        The computed mean tempo value.

    """

    return np.mean(beats_to_tempo(beats))


def _validate_beats(beats: Union[Iterable[float], np.ndarray]) -> None:
    """Checks an iterable of beat timestamps and raises errors as required"""

    if not isinstance(beats, np.ndarray):
        beats = np.array(beats, dtype=float)  # expecting a float dtype
    if np.isnan(np.sum(beats)):
        raise ValueError(
            "Cannot calculate instantaneous tempo measurements from array of beats with missing values"
        )
    if len(beats) < 2:
        raise ValueError("Must pass at least two beat timestamps")
    if len(beats.shape) != 1:
        raise ValueError("Beat timestamps must be 1-dimensional")
    if not np.array_equal(np.sort(beats), beats):
        raise ValueError("Beat timestamps must increase linearly")
    if len(np.unique(beats)) != len(beats):
        raise ValueError("Beat timestamps must not contain duplicates")


def beats_to_tempo(beats: np.ndarray) -> np.ndarray:
    """Converts beat timestamps to instantaneous tempo measurements."""

    # Raise error on invalid inputs
    _validate_beats(beats)
    return np.array(60 / np.diff(beats))


def fit_tempo_linear_regression(beats: Iterable[float]):
    """Fits linear regression of BPM measurements vs. onset time."""

    from scipy.stats import linregress

    # Target variable: BPM measurements
    # This will also validate the input
    beats_arr = np.array(beats)
    y = beats_to_tempo(beats_arr)
    # Predictor variable: the onset time
    x = beats_arr[1:]
    return linregress(x, y)



"""
Given a monotonically increasing series of integers (tatum indices),
process (and optionally visualise) GCD across all possible window sizes and start points.
"""

__author__ = "Mark Gotham"


from typing import Union

import matplotlib.colors as mpl
import matplotlib.pyplot as plt
import numpy as np

from amads.core.vector_transforms_checks import (
    indicator_to_interval,
    indices_to_indicator,
    is_monotonic,
)


def windowed_gcd(
    indices: list, exclude_redundant: bool = True, plot: bool = True
) -> Union[dict | None]:
    """
    Performs windowed GCD calculations on a monotonically increasing list of numbers
    and creates a visualization with time on the x-axis and window size on the y.

    Parameters
    ----------
    indices: A list of monotonically increasing numbers.
    exclude_redundant:
        If True, do not process or plot window sizes
        whose list of GCD is identical to the previous round (window size - 1).
        Default True to save on a lot of processing of trivial data.
    plot: Optionally run `windowed_gcd_visualization` to plot the data.

    Raises
    ------
    If the indices data is not monotonically increasing (see `is_monotonic`).

    Examples
    --------
    Internally, the function processes monotonic indices into windowed intervals.

    >>> monotonic_indices = [0, 1, 2, 4, 6, 7, 10]
    >>> indicator_vector = indices_to_indicator(monotonic_indices)
    >>> indicator_vector
    (1, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1)

    >>> indicator_to_interval(indicator_vector, wrap=False)
    (1, 1, 2, 2, 1, 3)

    This function returns a dict where
    keys are the window size,
    and for each key the value is a list of gcds for that size.
    We only store them all if `exclude_redundant=False`.

    >>> windowed_gcd_data = windowed_gcd(monotonic_indices, plot=False, exclude_redundant=False)

    >>> windowed_gcd_data.keys()
    dict_keys([2, 3, 4, 5, 6, 7, 8, 9, 10])

    By defaut, `exclude_redundant=True` so
    >>> windowed_gcd_data = windowed_gcd(monotonic_indices, plot=False)  # exclude_redundant=True
    >>> windowed_gcd_data.keys()
    dict_keys([2, 3, 4, 5, 6])

    >>> windowed_gcd_data[4]
    [1, 1, 2, 2, 1, 1, 1, 3]

    I.e., for the `monotonic_indices` [0, 1, 2, 4, 6, 7, 10] and window size 4 we have:

    0-3 -> indices 0, 1, 2 -> tatum 1
    1-4 -> indices 1, 2, 4 -> tatum 1
    2-5 -> indices 2, 4 -> tatum 2
    3-6 -> indices 4, 6 -> tatum 2
    4-7 -> indices 4, 6, 7 -> tatum 1
    5-8 -> indices 6, 7 -> tatum 1
    6-9 -> indices 6, 7 -> tatum 1
    7-10 -> indices 7, 10 -> tatum 3
    """
    is_monotonic(indices)  # Raises error if not

    indicator = indices_to_indicator(indices, indicator_length=max(indices) + 1)
    max_val = len(indicator)

    # Precompute once
    full_intervals = np.array(indicator_to_interval(indicator, wrap=False))
    ones_positions = np.where(np.array(indicator) == 1)[0]
    interval_starts = ones_positions[:-1]
    interval_ends = ones_positions[1:]

    gcds_by_window_size = {}
    max_gcd = 1
    last_gcd_list = []

    for window_size in range(2, max_val):

        i_values = np.arange(
            0, max_val - window_size + 1
        )  # shape: (n_windows,)

        # 2D boolean array of shape (n_windows, n_intervals).
        # Entry [i, j] is True if interval j fits entirely inside window i
        # (start is ≥ window start; end is < the window's end)
        mask = (interval_starts[np.newaxis, :] >= i_values[:, np.newaxis]) & (
            interval_ends[np.newaxis, :]
            < (i_values + window_size)[:, np.newaxis]
        )

        # Pad with 0 (identity for GCD) where interval is outside the window
        padded = np.where(
            mask, full_intervals[np.newaxis, :], 0
        )  # (n_windows, n_intervals)

        # Fully vectorised GCD across interval axis
        gcd_per_window = np.gcd.reduce(padded, axis=1)

        # Windows with no intervals in them (all-zero rows) should be sentinel 0
        gcd_per_window[~mask.any(axis=1)] = 0

        this_window_size_gcd_list = gcd_per_window.tolist()

        if exclude_redundant:

            if all(v == 0 for v in this_window_size_gcd_list):
                continue

            if this_window_size_gcd_list == last_gcd_list[:-1]:
                last_gcd_list = this_window_size_gcd_list
                continue

            last_gcd_list = this_window_size_gcd_list

        gcds_by_window_size[window_size] = this_window_size_gcd_list
        if max(this_window_size_gcd_list) > max_gcd:
            max_gcd = max(this_window_size_gcd_list)

    if plot:
        windowed_gcd_visualization(
            max_val=max_val,
            max_gcd=max_gcd,
            gcds_by_window_size=gcds_by_window_size,
        )
        return None
    else:
        return gcds_by_window_size


def windowed_gcd_visualization(
    max_val: int, max_gcd: int, gcds_by_window_size: dict
) -> None:
    """
    Create a triangular plot of windowed GCD values,
    analogous to the "keyscape" plots of Sapp et al. for the case of key.

    The x-axis represents tatum position, the y-axis represents window size,
    and each block is coloured by the GCD of the indices within that window.
    Larger window sizes appear higher on the plot, so the topmost row (if present)
    covers the entire piece — typically excluded as redundant by `windowed_gcd`.

    Parameters
    ----------
    max_val : int
        The length of the indicator vector, used to set the x-axis range.
    max_gcd : int
        The maximum GCD value across all windows, used to anchor the discrete colour map.
    gcds_by_window_size : dict
        A dict mapping window size (int) to a list of GCD values (one per tatum position),
        as returned by `windowed_gcd`.

    See Also
    --------
    windowed_gcd : Computes the `gcds_by_window_size` dict passed to this function.
    """
    fig, ax = plt.subplots(figsize=(12, 6))

    ax.grid(axis="x")
    ax.set_xlim(0, max_val)

    vmin, vmax = 0, max_gcd
    n_colors = vmax - vmin + 1

    cmap = plt.get_cmap("viridis", n_colors)
    bounds = np.arange(vmin - 0.5, vmax + 1.5, 1)
    norm = mpl.BoundaryNorm(bounds, cmap.N)
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm._A = []
    fig.colorbar(sm, ax=ax, ticks=range(vmin, vmax + 1))

    for window_size_key, gcd_list in gcds_by_window_size.items():
        for index, gcd in enumerate(gcd_list):
            color = cmap(norm(gcd))  # same norm for bars and colorbar
            ax.barh(window_size_key, 0.9, left=index, height=0.9, color=color)

    ax.set_xlabel("Tatum index")
    ax.set_ylabel("Window Size")
    plt.tight_layout()
    plt.show()


# ------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Number of notes per quarter or second in a Score.

Author:
    Tai Nakamura (2025)

Original Doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=77
"""

from typing import Optional

from amads.core.basics import Score

__author__ = "Tai Nakamura"


def note_density(score: Score, timetype: Optional[str] = "quarters") -> float:
    """
    Returns the number of notes per quarter or per second in a Score as a float.

    Specifically, compute note density as (number of notes - 1) divided by
    the time span from the first note onset to the last note onset.
    The subtraction of 1 ensures that density is measured in terms
    of intervals between notes.
    If there are no notes, return 0.0.

    <small>**Author**: Tai Nakamura</small>

    Parameters
    ----------
    score : Score
        The musical score to analyze.
    timetype : str, optional, default='quarters'
        Time unit for calculation:

        - 'quarters': notes per quarter (default)
        - 'seconds' : notes per second

    Returns
    -------
    float
        Computed note density. Returns 0.0 if the score
        is empty or if all notes have the same onset time
        (i.e., time span equals zero).

    Raises
    ------
    ValueError
        If `timetype` is not 'quarters' or 'seconds'.

    Examples
    --------
    >>> score = Score.from_melody([60, 62, 64, 65])  # all quarter notes
    >>> note_density(score, timetype='quarters')
    1.0
    >>> from amads.core.timemap import TimeMap
    >>> score = Score.from_melody([60, 62, 64, 65])  # all quarter notes
    >>> score.time_map = TimeMap(qpm = 120)  # set QPM to 120
    >>> note_density(score, timetype='seconds')
    2.0
    >>> score = Score.from_melody([60, 62, 64, 65], durations = [1.0, 2.0, 3.0, 4.0])  # mixed durations
    >>> note_density(score, timetype='quarters')
    0.5
    """
    notes = score.get_sorted_notes()
    if not notes:
        return 0.0

    if timetype == "seconds":
        if score.units_are_seconds:
            start_onset = notes[0].onset
            end_onset = notes[-1].onset
        else:
            start_onset = score.time_map.quarter_to_time(notes[0].onset)
            end_onset = score.time_map.quarter_to_time(notes[-1].onset)
    elif timetype == "quarters":
        if score.units_are_seconds:
            start_onset = score.time_map.time_to_quarter(notes[0].onset)
            end_onset = score.time_map.time_to_quarter(notes[-1].onset)
        else:
            start_onset = notes[0].onset
            end_onset = notes[-1].onset
    else:
        raise ValueError(
            f"Invalid timetype: {timetype}. Use 'quarters' or 'seconds'."
        )
    duration = end_onset - start_onset
    if duration <= 0:
        return 0.0
    return (len(notes) - 1) / duration



# example.py -- access to music examples
#
# Roger B. Dannenberg
# Sep 2024

import os
from importlib import resources, util

from amads.io.readscore import valid_score_extensions


def fullpath(example: str) -> str:
    """Construct a full path name for an example file.

    For example, fullpath("midi/sarabande.mid") returns a path to a
    readable file from this package.  This uses importlib so that
    we can read files even from compressed packages (we hope).

    Parameters
    ----------
    example : str
        The relative path to the example file, starting from the "music"
        directory. For example, "midi/sarabande.mid" or "musicxml/ex2.xml".

    Returns
    -------
    str
        The full path to the example file.

    Raises
    ------
    FileNotFoundError
        If the example file is not found or is not readable.
    """

    def trim_path(full: str) -> str:
        """remove first part of path to construct valid parameter value"""
        first_part = "amads/music/"
        index = full.find(first_part)
        return full if index == -1 else full[index + len(first_part) :]

    path = str(resources.files("amads").joinpath("music/" + example))

    if os.path.isfile(path) and os.access(path, os.R_OK):
        return path

    print("In amads.example.fullpath(" + example + "):")
    print("    File was not found. Try one of these:")

    spec = util.find_spec("amads")
    if spec is None:
        print("Error: Package amads not found")
        raise FileNotFoundError("Package amads not found")
    if spec.submodule_search_locations is None:
        print("Error: Package amads has no submodule search locations")
        raise FileNotFoundError(
            "Package amads has no submodule search locations"
        )
    package_path = spec.submodule_search_locations[0]

    # Walk through the directory hierarchy
    for root, dirs, files in os.walk(package_path):
        for file in files:
            for ext in valid_score_extensions:
                if file.endswith(ext):
                    parameter_option = trim_path(os.path.join(root, file))
                    print(f'   "{parameter_option}"')
    raise FileNotFoundError("Example file not found")



"""
Test suite for pcdist1
"""

from amads.core.basics import Score
from amads.pitch.pcdist1 import pitch_class_distribution_1


def test_empty_melody():
    score = Score.from_melody([])
    val = pitch_class_distribution_1(score)
    test = [0] * 12
    assert val.data == test



"""
This function implements the local boundary detection model
by Cambouropoulos (1997)

Function input:
    Musical score (as specified in basics.py)

Function output:
    (note start, strength) list
    note that since we are dealing with a monophonic score, each start
    corresponds to only 1 unique note, we can ensure that each output
    corresponds to a unique monophonic piece of music

Here's a few questions that we need to answer before deciding
on what our output should be...
(1) How do we represent causal relationship within our current score/part
structure?
(2) How do we encode additional information within the confines of the inherent
structure of the score? how do we do so in an extensible manner (preferably
external to the Score itself?)
Note that the matlab version has a column matrix of strengths, where the
row position of each weight corresponds to the row position of the note
in the score matrix...

What additional structure do we need to port the matlab version faithfully?
The problems are 2-fold:
(1) we have a partial ordering in our version of the score, instead of
the total ordering done in the matlab version.
(2) we don't have a way to correspond information to each score element
(e.g. note, measure, chord, etc.)

Things we need from basics.py:
(1) each event should have not only a parent, but also the ability to obtain
immediate successors and predecessors within the current containing object.

How about, we have a column matrix with weak references to the notes in the
argument score?
Note here that we can't segment the output into individual scores, because
there are no clear defined boundaries, only "soft" boundaries.
Another problem is, is it appropriate for each of these segmentation algorithms
to have wildly different output structures?
For that matter, what should we do with standardizing output structure?
Note to strength dictionary? Too complex.
Since we are only dealing with monophonic scores in these algorithms, we can
leverage the fact that each note start has a unique note in the score and
emit start and strength pairs...
"""

from amads.core.basics import Score


def boundary(score: Score) -> Score:
    """
    Compute the local boundary strength with LBDM (Cambouropoulos, 1997).

    Raises
    ------
    ValueError
        if the score is not monophonic

    Returns
    -------
    Score
        where the notes have the "boundary_strength" property set to a
        value from 0 to 1.
    """
    if not score.ismonophonic():
        raise ValueError("Score must be monophonic")

    notes = score.get_sorted_notes()

    # profiles
    pp = [
        abs(pair[1].key_num - pair[0].key_num) for pair in zip(notes, notes[1:])
    ]
    po = [pair[1].onset - pair[0].onset for pair in zip(notes, notes[1:])]
    pr = [
        max(0, pair[1].onset - pair[0].offset) for pair in zip(notes, notes[1:])
    ]

    def list_degrees(profile):
        ret_list = [
            abs(pair[1] - pair[0]) / (1e-6 + pair[1] + pair[0])
            for pair in zip(profile, profile[1:])
        ]
        ret_list.append(0)
        return ret_list

    # degrees of change
    rp = list_degrees(pp)
    ro = list_degrees(po)
    rr = list_degrees(pr)

    def list_strengths(profile, degrees):
        degrees_sum = [0]
        for degree_pair in zip(degrees, degrees[1:]):
            degrees_sum.append(degree_pair[0] + degree_pair[1])
        strengths = [pair[0] * pair[1] for pair in zip(profile, degrees_sum)]
        max_strength = max(strengths)
        if max_strength > 0.1:
            strengths = [x / max_strength for x in strengths]
        return strengths

    sp = list_strengths(pp, rp)
    so = list_strengths(po, ro)
    sr = list_strengths(pr, rr)

    if len(notes) > 0:
        notes[0].set("boundary_strength", 1)
    for sp_elem, so_elem, sr_elem, note in zip(sp, so, sr, notes[1:]):
        note.set(
            "boundary_strength", 0.25 * sp_elem + 0.5 * so_elem + 0.25 * sr_elem
        )

        # b = [1]
        # for sp_elem, so_elem, sr_elem in zip(sp, so, sr):
        #    b.append(0.25 * sp_elem + 0.5 * so_elem + 0.25 * sr_elem)
    # assert len(b) == len(notes)

    return score



# m21_show.py - print Music21 score content

from typing import Union

from music21 import metadata, stream


def music21_show(
    m21score: Union[stream.Score, stream.Part], filename: str
) -> None:
    """print Music21 score structure"""
    print("Music21 score structure", end="")
    if filename is not None:
        print(f" for {filename}:")
    else:
        print(":")
    for element in m21score:
        if isinstance(element, metadata.Metadata):
            print(element.all())

    # Print tempo changes
    print("\nTempo changes:")
    tempos = m21score.flatten().getElementsByClass("MetronomeMark")
    if tempos:
        for tempo in tempos:
            print(f"  Offset {tempo.offset}: {tempo.number} BPM")
    else:
        print("  No explicit tempo marks found")

    # Print elements
    print(m21score.show("text", addEndTimes=True))



import warnings
from math import isclose
from typing import Dict, Optional, Tuple, Union, cast

from music21 import (
    bar,
    chord,
    clef,
    converter,
    instrument,
    key,
    note,
    stream,
    tempo,
)
from music21.meter.base import TimeSignature as m21TimeSignature

from amads.core.basics import (
    Chord,
    Clef,
    KeySignature,
    Measure,
    Note,
    Part,
    Pitch,
    Rest,
    Score,
    Staff,
    TimeSignature,
)
from amads.io.m21_show import music21_show

tied_notes = {}  # temporary data to track tied notes, this is a mapping
# from key number to Note object for notes that originate a tie. When we
# see a note that ends or continues a tie, we look up the origin of the
# tie in this dictionary and link it to the note that ends or continues
# the tie. Note that we do not encode 'let-ring' or 'continue-let-ring'
# in the AMADS data model, so we ignore those cases.
#     It is possible for notes to overlap in time and be tied. What ties
# to what in Music21 (as well as MIDI files) is ambiguous in this case,
# but ignoring overlapping ties would create multiple notes when there
# was only one in the MIDI file, so when there is overlap, we map from
# pitch to a list of notes that originate unterminated ties for this
# pitch. When a tie is terminated, we tie from the first note in the
# list. (First-in-first-out).


def music21_import(
    filename: str,
    format: str,
    flatten: bool = False,
    collapse: bool = False,
    show: bool = False,
    group_by_instrument: bool = True,
) -> Score:
    """
    Use music21 to import a file and convert it to a Score.

    Parameters
    ----------
    filename : str
        The path to the music file.
    format: str
        The file format: 'musicxml', 'kern', 'mei'
    flatten : bool, optional
        If True, flatten the score structure.
    collapse : bool, optional
        If True and flatten is true, also collapse parts.
    show : bool, optional
        If True, print the music21 score structure for debugging.
    group_by_instrument : bool, optional
        If True, group parts by instrument name into staffs. Defaults to True.
        See music21_to_score() for more details.

    Returns
    -------
    Score
        The converted AMADS Score object.
    """
    # Load the file using music21
    if format == "kern":
        format = "humdrum"  # music21 uses "humdrum" for kern files
    # other formats are the same for both AMADS and music21
    qp = format != "midi"  # quantize non-MIDI files until we discover this is
    # a bad idea. MIDI files may have expressive timing, so we never quantize.
    m21score = converter.parse(
        filename, format=format, forceSource=True, quantizePost=qp
    )

    # m21score can be an Opus, but this is checked in music21_to_score, so we
    # can ignore the type error here:
    score = music21_to_score(
        m21score,  # type: ignore
        flatten,
        collapse,
        show,  # type: ignore
        filename,
        group_by_instrument=group_by_instrument,
    )
    return score


# mapping from PartStaff id to list of AMADS Part, where each Part contains
# a Staff. After everything is read, Parts with the same id should be
# combined.
_staff_id_to_part: Dict[str, Tuple[stream.PartStaff, Part]]  # declare global


def music21_to_score(
    m21score: Union[stream.Score, stream.Part],
    flatten: bool = False,
    collapse: bool = False,
    show: bool = False,
    filename: Optional[str] = None,
    group_by_instrument: bool = True,
) -> Score:
    global _staff_id_to_part
    _staff_id_to_part = {}

    if show:
        music21_show(m21score, filename)  # type: ignore

    # Create an empty Score object
    duration = m21score.duration.quarterLength
    score = Score(duration=float(duration))

    parts: list[Part | None] = []  # All the Parts in score
    m21parts: list[stream.Part] = []  # Corresponding m21 parts
    group_ids: list[int] = []  # list of group identifiers
    next_group_id = 0  # groups are named by small integers
    if isinstance(m21score, stream.Score):
        group_ids = [-1] * len(m21score.parts)

    # Iterate over parts in the music21 score

    # We start with separate Part with Staff for each original part or staff.
    # We also have three lists that are "parallel" with Parts:
    # parts[i] is a list of all the Parts in order
    # m21parts[i] is the list of m21 parts corresponding to parts[i]
    # group_ids[i] is the group that the ith Part belongs to, so we can search
    #     to find all Parts whose Staff belong together.

    if isinstance(m21score, stream.Part):
        music21_convert_part(m21score, score, duration)
    elif isinstance(m21score, stream.Score):
        shared_shift = None  # used to check all parts have same time shift
        for i, m21part in enumerate(m21score.parts):
            if isinstance(m21part, stream.Part):
                # Convert the music21 part into an AMADS Part and
                # append it to the Score:
                part, shift = music21_convert_part(m21part, score, duration)
                # check that all parts have the same time shift; this should
                # be true if all first measures have the same number of beats
                # either as notes or rest or some combination.
                if shared_shift is not None:
                    assert isclose(shared_shift, shift, abs_tol=0.001)
                else:
                    shared_shift = shift
                parts.append(part)
                m21parts.append(m21part)
                # if m21part is a PartStaff (subclass of Part), keep
                # a dictionary of names and Parts:
                if isinstance(m21part, stream.PartStaff):
                    next_group_id += 1
                    group_id = next_group_id
                    spanners = m21part.getSpannerSites()
                    for spanner in spanners:
                        if spanner.__class__.__name__ == "StaffGroup":
                            for elem in spanner.getSpannedElements():
                                if isinstance(elem, stream.PartStaff):
                                    assert elem in m21parts
                                    j = m21parts.index(elem)
                                    group_ids[j] = group_id
                                    break  # only one PartStaff spanner
            else:
                warnings.warn(
                    f"Ignoring non-Part element of Music21 score: {m21part}"
                )
    else:
        raise ValueError("expected Score or Part from music21 reader")

    # If group_by_instrument, we form groups by looking for matching
    # instruments:
    if group_by_instrument:
        instruments = []
        for i in range(len(parts)):
            try:
                # later, parts[i] can be None, so type checker complains here:
                j = instruments.index(parts[i].instrument)  # type: ignore
            except ValueError:
                instruments.append(parts[i].instrument)  # type: ignore
                continue
            instruments.append(parts[i].instrument)  # type: ignore
            if group_ids[i] == -1:  # part j is not in a group
                if group_ids[j] == -1:  # neither is i, so make a new group
                    next_group_id += 1
                    group_ids[j] = next_group_id
                    group_ids[i] = next_group_id
                    next_group_id += 1
                else:  # j is in a group, so assign that to i
                    group_ids[i] = group_ids[j]
            # else both i and j are in groups already, must be from MusicXML
            # staff grouping; probably correct already, so do no merge them

    # Now, grouping is complete.
    # To process Parts, for each i, see if group_ids[i] is not None. If not,
    # find all other Parts with matching group_id and merge them. As you do it,
    # remove the other Parts from the score.

    for i in range(len(parts)):
        to_part = cast(Part, parts[i])
        if group_ids[i] != -1:
            # find all other parts with matching group_ids[j]:
            group_id = group_ids[i]
            j = i
            while True:
                group_ids[j] = -1  # so j will not match
                try:
                    j = group_ids.index(group_id)
                except ValueError:  # no match, nothing else in group
                    break  # no more staffs to merge
                from_part = cast(Part, parts[j])
                parts[j] = None  # soon this part will be gone
                if to_part.instrument is None:
                    to_part.instrument = from_part.instrument
                score.remove(from_part)
                from_staff = cast(Staff, from_part.content[0])
                from_part.remove(from_staff)
                to_part.insert(from_staff)

    if flatten or collapse:
        score = score.flatten(collapse=collapse)
    return score


def music21_convert_note(m21note, measure):
    """
    Convert a music21 note into an AMADS Note and append it to the Measure.

    Parameters
    ----------
    m21note : music21.note.Note
        The music21 note to convert.
    measure : Measure
        The Measure object to which the converted Note will be appended.
    """
    duration = float(m21note.duration.quarterLength)
    dynamic = m21note.volume.velocity
    if isinstance(dynamic, int):
        dynamic = min(max(dynamic, 1), 127)
    note = Note(
        parent=measure,
        onset=float(measure.onset + m21note.offset),
        pitch=Pitch(pitch=m21note.pitch.midi, alt=m21note.pitch.alter),
        duration=duration,
        dynamic=dynamic,
    )
    if m21note.tie is not None:
        music21_convert_tie(m21note.pitch.midi, note, m21note.tie.type)


def music21_convert_tie(key_num: int, note: Note, tie_type: str) -> None:
    """Handle tie to and/or from music21 note

    Parameters
    ----------
    key_num: int
        the MIDI key number (pitch)
    note : Note
        the note we are creating, corresponds to m21note
    tie_type : str
        the tie type, one of "start", "continue", "stop"
    """
    if tie_type == "start":
        # Start of a tie
        if key_num in tied_notes:
            # If the note is already tied, we should not see "start":
            import warnings

            warnings.warn(
                f"music21 note (key_num {key_num} at beat"
                f" {note.onset}) starts a tie, but there is already"
                " an open tie for that pitch. Maybe MIDI file has"
                " multiple note-on events without an intervening"
                " note-off event."
            )
            # make a list of started ties
            tied_note = tied_notes[key_num]
            if isinstance(tied_note, list):
                tied_note.append(note)
            else:
                tied_notes[key_num] = [tied_note, note]
        else:
            tied_notes[key_num] = note
    elif tie_type == "continue":  # Continuation of a tie
        if key_num in tied_notes:
            origin = tied_notes[key_num]
            if isinstance(origin, list):
                origin_note = origin.pop(0)
                origin_note.tie = note
                origin.append(note)  # this note is tied to something too
                origin = origin_note
            else:
                tied_notes[key_num] = note  # to be continued :-)
            origin.tie = note
        else:  # missing start note
            import warnings

            warnings.warn(
                f"music21 note (key_num {key_num} at beat"
                f" {note.onset}) continues a tie, but there is no"
                " start note for that pitch."
            )
    elif tie_type == "stop":  # End of a tie
        if key_num in tied_notes:
            origin = tied_notes[key_num]
            if isinstance(origin, list):
                origin_note = origin.pop(0)
                if len(origin) == 1:  # restore to non-list single note
                    tied_notes[key_num] = origin[0]
                origin = origin_note
            else:
                del tied_notes[key_num]  # remove the origin
            origin.tie = note
        else:  # missing start note
            import warnings

            warnings.warn(
                f"music21 note (key_num {key_num} at beat"
                f" {note.onset}) ends a tie, but there is no start"
                " note for that pitch."
            )


def music21_convert_rest(m21rest, measure):
    """
    Convert a music21 rest into an AMADS Rest and append it to the Measure.

    Parameters
    ----------
    m21rest : music21.note.Rest
        The music21 rest to convert.
    measure : Measure
        The Measure object to which the converted Rest will be appended.
    """
    duration = float(m21rest.quarterLength)
    # Create a new Rest object and associate it with the Measure
    Rest(
        parent=measure,
        onset=float(measure.onset + m21rest.offset),
        duration=duration,
    )


def music21_convert_chord(m21chord, measure, offset):
    """
    Convert a music21 chord into an AMADS Chord and append it to the Measure.
    Apparently, chord notes cannot be tied, so we ignore ties.

    Parameters
    ----------
    m21chord : music21.chord.Chord
        The music21 chord to convert.
    measure : Measure
        The Measure object to which the converted Chord will be appended.
    """
    duration = float(m21chord.quarterLength)
    chord = Chord(
        parent=measure,
        onset=float(measure.onset + m21chord.offset),
        duration=duration,
    )
    for pitch in m21chord.pitches:
        note = Note(
            parent=chord,
            onset=float(measure.onset + m21chord.offset),
            pitch=Pitch(pitch=pitch.midi, alt=pitch.alter),
            duration=duration,
        )
        if m21chord.tie is not None:
            music21_convert_tie(pitch.midi, note, m21chord.tie.type)


def update_part_instrument(caller_id, part, m21instr):
    assert part is not None
    name = m21instr.partName
    name = None if name == "Unknown" else name
    if part.instrument is None:
        part.instrument = name
    elif name != part.instrument:
        warnings.warn(
            f"Music21_convert_measure ignoring {m21instr.__class__}"
            f" ({m21instr}) because part already has different"
            f" instrument {part.instrument}."
        )
        part_program = part.get("midi_program")
        measure_program = m21instr.midiProgram
        if part_program is None:
            part.set("midi_program", measure_program)
        elif part_program != measure_program:
            warnings.warn(
                f"music21.instrument midi_program conflict: {m21instr},"
                " program change ignored)"
            )


def _remove_clef_from_measure(measure: Measure, onset: float) -> None:
    """Removes Clefs near onset time in measure."""
    to_remove = [
        elem
        for elem in measure.content
        if (
            isinstance(elem, Clef) and isclose(elem.onset, onset, abs_tol=0.001)
        )
    ]
    for elem in to_remove:
        measure.remove(elem)


def append_items_to_measure(
    measure: Measure, source: stream.Stream, offset: float
) -> None:
    """
    Append items from a source to the Measure.

    Parameters
    ----------
    measure : Measure
        The Measure object to which items will be appended.
    source : music21.stream.Stream
        The source stream containing items to append.
    """
    for element in source.iter():
        if isinstance(element, note.Note):
            music21_convert_note(element, measure)
        elif isinstance(element, note.Rest):
            music21_convert_rest(element, measure)
        elif isinstance(element, m21TimeSignature):
            # Create a TimeSignature object and insert into the score
            # if the TimeSignature changes at this time:
            upper = element.numerator
            lower = element.denominator
            if abs(element.offset) > 1e-3:
                warnings.warn(
                    "Music21 time signature found that is not at the"
                    f" measure beginning: {element}. Moving the"
                    " signature to the beginning of the measure."
                )
            # what TimeSignature is in effect?
            ts = measure.time_signature()
            if ts.upper != upper or ts.lower != lower:
                last_ts = measure.score.time_signatures[-1]  # type: ignore
                if last_ts.time > measure.onset:
                    warnings.warn(
                        "Encountered a new Music21 time signature"
                        " placed BEFORE an earlier time signature:"
                        " {element}. Something is probably wrong"
                        " with this score. Ignoring the time"
                        " signature in conversion to AMADS."
                    )
                else:
                    ts = TimeSignature(measure.onset, upper, lower)
                    measure.score.append_time_signature(ts)  # type: ignore
        elif isinstance(element, key.KeySignature):
            # Create a KeySignature object and associate it with the Measure
            KeySignature(
                measure, measure.onset + element.offset, key_sig=element.sharps
            )
        elif isinstance(element, clef.Clef):
            # Partitura will write Clef for staff 1 without a number attribute
            # in MusicXML, and if there is a <clef> and a <clef number="2">,
            # Music21 apparently puts *both* into staff 2. To prevent this,
            # we will remove any Clef appearing at the same time in the same
            # measure
            _remove_clef_from_measure(measure, measure.onset + element.offset)
            # Create a Clef object and associate it with the Measure
            Clef(measure, measure.onset + element.offset, clef=element.name)
        elif isinstance(element, chord.Chord):
            music21_convert_chord(element, measure, offset)
        elif isinstance(element, stream.Voice):
            # Voice containers are ignored, so promote contents to the Measure
            append_items_to_measure(measure, element, offset + element.offset)
        elif isinstance(element, tempo.MetronomeMark):
            # update tempo
            time_map = measure.score.time_map  # type: ignore (measure has
            #     a parent)
            last_beat = time_map.changes[-1].quarter
            tempo_change_onset = offset + element.offset
            if last_beat > tempo_change_onset:
                warnings.warn(
                    f"music21 tempo mark at {tempo_change_onset}"
                    " is within existing time mmap, ignoring"
                )
            else:
                qpm = element.getQuarterBPM()
                # ignore music21 tempo mark if it returns None for BPM
                if qpm is None:
                    warnings.warn(
                        f"Music21 tempo mark at {tempo_change_onset}"
                        " has no BPM, ignoring"
                    )
                else:
                    time_map.append_change(tempo_change_onset, qpm)
        elif isinstance(element, bar.Barline):
            pass  # ignore barlines, e.g. Barline type="final"
        elif isinstance(element, instrument.Instrument):
            part = measure.part
            update_part_instrument("FOUND INSTRUMENT IN MEASURE", part, element)
        else:
            warnings.warn(
                "Music21_convert_measure ignoring non-Note element"
                f" {element} : {element.__class__}."
            )


def music21_convert_measure(m21measure, staff):
    """
    Convert a music21 measure into an AMADS Measure and append it to the Staff.

    Parameters
    ----------
    m21measure : music21.stream.Measure
        The music21 measure to convert.
    staff : Staff
        The Staff object to which the converted Measure will be appended.
    """
    # Create a new Measure object and associate it with the Staff
    measure = Measure(
        parent=staff,
        onset=m21measure.offset,
        duration=float(m21measure.barDuration.quarterLength),
    )

    # Iterate over elements in the music21 measure
    append_items_to_measure(measure, m21measure, m21measure.offset)
    return measure


def music21_convert_part(m21part, score, duration):
    """
    Convert a music21 part into an AMADS Part and append it to the Score.

    Parameters
    ----------
    m21part : music21.stream.Part
        The music21 part to convert.
    score : Score
        The Score object to which the converted Part will be appended.
    """
    global tied_notes  # temporary data to track tied notes
    # Create a new Part object and associate it with the Score
    name = m21part.partName
    if name == "Unknown":  # AMADS uses "Unknown" to represent None.
        # Of course, if a user really names an instrument "Unknown",
        # that particular name will not be stored in the Part.
        name = None
    part = Part(parent=score, instrument=name, duration=duration)
    staff = Staff(parent=part)  # Assuming a single staff for simplicity
    tied_notes.clear()
    # Iterate over elements in the music21 part
    for element in m21part.iter():
        if isinstance(element, stream.Measure):
            # Convert music21 Measure to our Measure class
            music21_convert_measure(element, staff)
        elif isinstance(element, instrument.Instrument):
            update_part_instrument("FOUND INSTRUMENT IN PART", part, element)
        else:
            warnings.warn(
                f"music21_convert_part ignoring non-Measure element: {element}"
            )
    if len(tied_notes.keys()) > 0:
        warnings.warn(
            f"music21_convert_part: tied notes in {part} from these"
            f" notes were not closed at the end of the part:"
            f" {tied_notes.values()}"
        )
    tied_notes.clear()
    staff.offset = staff.content[-1].offset

    shift = 0  # how much did we shift to make a full measure 1?

    # expand first measure to a full measure if necessary
    # what is the maximum offset of the first measure?
    if len(staff.content) > 0:
        m1 = staff.content[0]
        m1 = cast(Measure, m1)
        max_offset = 0
        for elem in m1.content:
            max_offset = max(max_offset, elem.offset)
        if max_offset < m1.offset - 0.001:  # need to insert rest
            shift = m1.offset - max_offset
            for elem in m1.content:
                if (
                    isinstance(elem, Note)
                    or isinstance(elem, Rest)
                    or isinstance(elem, Chord)
                ):
                    elem.time_shift(shift)
            # insert Rest, Since m1 is first, m1.onset == 0
            _ = Rest(m1, m1.onset, duration=shift)

            # now, first measure ending may have shifted, so adjust
            # remainder of the part
            if len(staff.content) > 1:
                m2 = staff.content[1]
                m2 = cast(Measure, m2)
                shift = m1.offset - m2.onset
                if shift > 0.001:  # need to shift everything
                    for m in staff.content[1:]:
                        m.time_shift(shift)
            staff.offset = staff.content[-1].offset
            part.offset = max(part.offset, staff.offset)
            score.offset = max(score.duration, staff.offset)
    return part, shift



"""
Functions for retrieving one pitch class set property directly from another.

Most of the retrieval function names are in the form
`<call-on-type>_to_<return-type>`
e.g.
`prime_to_combinatoriality`
Some are simple mappings from one entry to another.
Anything starting with pitches involves more calculation.

Broadly as implemented by Mark Gotham
for [Serial_Analyser](https://github.com/MarkGotham/Serial_Analyser)
after Robert Morris,
but the functions are general
and not really named algorithms in sense used elsewhere on this code base.

<small>**Author**: Mark Gotham</small>
"""

__author__ = "Mark Gotham"

from typing import Sequence

from amads.core.vectors_sets import multiset_to_vector, pairwise_differences
from amads.pitch import pc_sets
from amads.pitch import transformations as pitch_list_transformations


def set_classes_from_cardinality(cardinality: int) -> list:
    """
    Find pitch class set data matching given cardinality.

    Parameters
    ----------
    cardinality: int
        The cardinality of the set (2–10 inclusive).

    Returns
    -------
    list
        The pitch class set data for that cardinality.

    Examples
    --------
    >>> set_classes_from_cardinality(2)[0]
    ('2-1', (0, 1), (1, 0, 0, 0, 0, 0), 2)

    """
    if not (1 < cardinality < 11):
        raise ValueError("Invalid cardinality: must be 2-10 (inclusive).")
    return pc_sets.set_classes[cardinality]


def prime_to_combinatoriality(prime: tuple[int, ...]) -> int:
    """
    Find the combinatoriality status for a given prime form.

    Parameters
    ----------
    prime: tuple[int, ...]
        A prime form expressed as a tuple of integers.

    Returns
    -------
    int
        The number of distinct transformations (non-invariant transpositions and / or inversions).

    Examples
    --------
    >>> prime_to_combinatoriality((0, 1, 2, 3))
    12
    """
    data = set_classes_from_cardinality(len(prime))
    for x in data:
        if x[1] == prime:
            return x[3]
    raise ValueError(f"{prime} is not a valid prime form")


def interval_vector_to_combinatoriality(vector: tuple[int, ...]) -> str:
    """
    Find the combinatoriality status for a given interval vector.

    Parameters
    ----------
    vector: tuple[int, ...]
        An interval vector for any set with 2–10 distinct pitches,
        expressed as a tuple of 6 integers.

    Returns
    -------
    str
        The combinatoriality status: one of ``T``, ``I``, ``RI``, ``A``,
        or an empty string for non-combinatorial cases.

    Examples
    --------
    >>> interval_vector_to_combinatoriality((4, 3, 2, 1, 0, 0))  # 5-1
    12
    """
    if len(vector) != 6:
        raise ValueError(f"{vector} is not a valid interval vector")
    total = sum(vector)
    total_to_cardinality = {
        1: 2,
        3: 3,
        6: 4,
        10: 5,
        15: 6,
        21: 7,
        28: 8,
        36: 9,
        45: 10,
    }
    if total not in total_to_cardinality:
        raise ValueError(f"{vector} is not a valid interval vector")
    data = set_classes_from_cardinality(total_to_cardinality[total])
    for x in data:
        if x[2] == vector:
            return x[-1]
    raise ValueError(f"{vector} is not a valid interval vector")


def interval_to_interval_class(interval: int) -> int:
    """
    Map an interval (any integer, positive or negative and any size)
    to an interval class (integer in the range 0–6).

    Parameters
    ----------
    interval: int
        Any integer representing a pitch interval.

    Returns
    -------
    int
        The interval class in the range 0–6.

    Examples
    --------
    >>> interval_to_interval_class(0)
    0

    >>> interval_to_interval_class(-1)
    1

    >>> interval_to_interval_class(-2)
    2

    >>> interval_to_interval_class(11)
    1

    >>> interval_to_interval_class(7)
    5

    >>> interval_to_interval_class(-100)
    4
    """
    reduced = abs(interval) % 12
    return reduced if reduced <= 6 else 12 - reduced


def interval_vector_to_interval_class_vector(
    interval_vector: tuple[int, ...]
) -> tuple[int, ...]:
    """
    Map an interval vector of range(0, 12) to an interval class vector of range(1, 7).

    Parameters
    ----------
    interval_vector: tuple[int, ...]
        A 12-element interval vector.

    Returns
    -------
    tuple[int, ...]
        A 6-element interval class vector.

    Examples
    --------
    >>> interval_vector_to_interval_class_vector((1, 0, 5, 7, 2, 2, 0, 3, 5, 0, 8, 4))
    (4, 13, 7, 7, 5, 0)
    """
    interval_class_vector = [0] * 6
    for i in range(1, 6):
        interval_class_vector[i - 1] = (
            interval_vector[i] + interval_vector[12 - i]
        )
    interval_class_vector[5] = interval_vector[6]  # special case
    return tuple(interval_class_vector)


def pitches_to_combinatoriality(pitches: Sequence[int]) -> str:
    """
    Find the combinatoriality status for a given list of pitches.

    Parameters
    ----------
    pitches: Sequence[int]
        A sequence of integers (0–11) for sets with 2–10 distinct pitches.

    Returns
    -------
    str
        The combinatoriality status as a string.

    Examples
    --------
    >>> pitches_to_combinatoriality((0, 1, 2, 3, 4, 5, 6, 7, 8, 9))
    12
    """
    icv = set_to_interval_vector(pitches)
    return interval_vector_to_combinatoriality(icv)


def distinct_pcs(pitches: Sequence[int]) -> list[int]:
    """
    Find the distinct pitch classes for a given sequence of pitches.

    Parameters
    ----------
    pitches: Sequence[int]
        A sequence of pitches (any integers).

    Returns
    -------
    list[int]
        A list of distinct pitch classes in the range 0–11.

    Examples
    --------
    >>> distinct_pcs((0, 0, 1, 1, 14))
    [0, 1, 2]
    """
    return list({p % 12 for p in pitches})


def set_to_interval_vector(
    user_set: Sequence[int],
) -> tuple[int, ...]:
    """
    Generates the interval vector for a set of pitches:
    the counts of each interval class present between all pairs of elements.

    Parameters
    ----------
    user_set: Sequence[int]
        A sequence of integers representing pitch classes.
        Order does not affect the result.

    Returns
    -------
    tuple[int, ...]
        A tuple of length 6 where index ``i`` contains the count of
        interval class ``i+1`` among all pairs in ``user_set``.

    Examples
    --------
    >>> set_to_interval_vector([1, 3, 4, 6])
    (1, 2, 2, 0, 1, 0)

    >>> set_to_interval_vector([0, 4, 7])  # C major triad
    (0, 0, 1, 1, 1, 0)

    """
    differences = [
        interval_to_interval_class(x) for x in pairwise_differences(user_set)
    ]
    differences = multiset_to_vector(differences, max_index=6)
    return tuple(differences[1:])


def pitches_to_forte_class(pitches: Sequence[int]) -> str:
    """
    Find the Forte class for a given list of pitches.

    Parameters
    ----------
    pitches: Sequence[int]
        A sequence of integers (0–11) for sets with 2–10 distinct pitches.

    Returns
    -------
    str
        The Forte class.

    Examples
    --------
    >>> pitches_to_forte_class((0, 1, 2, 3, 4, 5, 6, 7, 8))
    '9-1'
    """
    data = set_classes_from_cardinality(len(pitches))
    prime = pitches_to_prime(pitches)
    for x in data:
        if x[1] == prime:
            return x[0]
    raise ValueError(f"{pitches} is not a valid entry.")


def pitches_to_prime(pitches: Sequence[int]) -> tuple[int, ...]:
    """
    Find the prime form for a given list of pitches.

    This function first converts the pitches to their interval vector.
    (That step is easy and fast).
    This vector can then unambiguously give the prime form except for Z-related pairs.
    This affects one pair of tetrachords (2 prime forms) and 15 pairs of hexachords (30 primes).
    In those cases, the prime form is worked out by comparing the pitch list against
    the pair of options in both inversions until a match is found.

    Parameters
    ----------
    pitches: Sequence[int]
        A sequence of integers (0–11) for sets with 2–10 distinct pitches.

    Returns
    -------
    tuple[int, ...]
        The prime form.

    Examples
    --------
    >>> pitches_to_prime((9, 0, 1, 2, 3, 4, 5, 6, 7, 8))
    (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
    """
    pitches = distinct_pcs(pitches)
    vector = set_to_interval_vector(pitches)
    data = set_classes_from_cardinality(len(pitches))

    primes = [x[1] for x in data if x[2] == vector]

    if len(primes) == 1:
        return primes[0]
    elif len(primes) > 1:
        for prime in primes:
            inverted = pitch_list_transformations.invert(prime)
            for t in [prime, inverted]:
                if transposition_equivalent(t, pitches):
                    return prime
    raise ValueError(f"{pitches} did not match any prime form.")


def transposition_equivalent(
    set1: Sequence[int],
    set2: Sequence[int],
) -> bool:
    """
    Determine whether two pitch class sets are transposition equivalent.

    Parameters
    ----------
    set1: Sequence[int]
        A pitch class set as a sequence of integers (0–11).
    set2: Sequence[int]
        A pitch class set as a sequence of integers (0–11).

    Returns
    -------
    bool
        True if the two sets are transposition equivalent, False otherwise.

    Examples
    --------

    Meaningully True:
    >>> transposition_equivalent((0, 1, 2, 3, 4, 5, 6, 7, 8), (1, 2, 3, 4, 5, 6, 7, 8, 9))
    True

    Trivially True:
    >>> transposition_equivalent((0, 1, 2, 3, 4, 5, 6, 7, 8), (0, 1, 2, 3, 4, 5, 6, 7, 8))
    True

    False:
    >>> transposition_equivalent((0, 1, 2, 3, 4, 5, 6, 7, 8), (0, 1, 2, 3, 4, 5, 6, 7))
    False

    """
    sorted_set2 = sorted(set2)
    for i in range(12):
        if (
            sorted(pitch_list_transformations.transpose_by(set1, i))
            == sorted_set2
        ):
            return True
    return False


# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



""" test_timemap.py -- tests for timemap class
"""

import pytest

from amads.core.timemap import TimeMap


def test_timemap_deepcopy():
    """Test that deepcopying a TimeMap works as expected."""
    import copy

    tm1 = TimeMap()
    tm1.append_change(0.0, 100)
    tm1.append_change(1.0, 120)
    tm1.append_change(2.0, 140)

    tm1.show()

    tm2 = copy.deepcopy(tm1)
    tm2.show()

    assert tm1 is not tm2
    for q1, q2 in zip(tm1.changes, tm2.changes):
        assert q1 is not q2
        assert q1.time == q2.time
        assert q1.quarter == q2.quarter


def test_timemap_mapping():
    """Test quarter_to_tempo method of TimeMap."""
    tm1 = TimeMap()
    tm1.append_change(0.0, 100)
    tm1.append_change(1.0, 120)
    tm1.append_change(2.0, 140)

    assert tm1.quarter_to_tempo(0) == pytest.approx(100)
    assert tm1.quarter_to_tempo(0.5) == pytest.approx(100)
    assert tm1.quarter_to_tempo(1) == pytest.approx(120)
    assert tm1.quarter_to_tempo(1.5) == pytest.approx(120)
    assert tm1.quarter_to_tempo(2) == pytest.approx(140)
    assert tm1.quarter_to_tempo(3) == pytest.approx(140)

    assert tm1.time_to_tempo(0) == pytest.approx(100)
    assert tm1.time_to_tempo(0.5 * 60 / 100) == pytest.approx(100)
    assert tm1.time_to_tempo(1 * 60 / 100) == pytest.approx(120)
    assert tm1.time_to_tempo(1 * 60 / 100 + 0.5 * 60 / 120) == pytest.approx(
        120
    )
    assert tm1.time_to_tempo(1 * 60 / 100 + 1 * 60 / 120) == pytest.approx(140)
    assert tm1.time_to_tempo(10) == pytest.approx(140)



"""
Basic tests of rhythmic/metrical profiles and syncopation metrics
exactly as reported in named Figures of
Gomez et al. (the WNBD paper)
and
Toussaint 2013.

We keep these tests separate for ease of cross-reference with the source,
preserving even the source order and broad structure.

"""

__author__ = "Mark Gotham"

from fractions import Fraction

from amads.core.vector_transforms_checks import (
    indicator_to_interval,
    is_maximally_even,
)
from amads.time.meter import profiles, syncopation
from amads.time.rhythm import (
    has_deep_property,
    has_oddity_property,
    keith_via_toussaint,
    off_beatness,
    vector_to_onset_beat,
)
from amads.time.variability import normalized_pairwise_variability_index

# 16
SHIKO = profiles.WorldSample16.shiko
SON = profiles.WorldSample16.son
SOUKOUS = profiles.WorldSample16.soukous
RUMBA = profiles.WorldSample16.rumba
BOSSA = profiles.WorldSample16.bossa_nova
GAHU = profiles.WorldSample16.gahu

# 12
SOLI = profiles.WorldSample12.soli
TAMBU = profiles.WorldSample12.tambú
BEMBE = profiles.WorldSample12.bembé
BEMBE_2 = profiles.WorldSample12.bembé_2
YORUBA = profiles.WorldSample12.yoruba
TONADA = profiles.WorldSample12.tonada
ASAADUA = profiles.WorldSample12.asaadua
SORSONET = profiles.WorldSample12.sorsonet
BEMBA = profiles.WorldSample12.bemba
ASHANTI = profiles.WorldSample12.ashanti


def count_non_zero(data):
    count = 0
    for element in data:
        if element != 0:
            count += 1
    return count


def test_profile_basics():
    for profile in (SHIKO, SON, SOUKOUS, RUMBA, BOSSA, GAHU):
        assert profile.count(1) == 5  # Onsets
        assert len(profile) == 16  # Grid


def test_toussaint_17_13():
    """
    Test the 16-unit rhythms with a beat pattern every 4th element
    against "Figure" 17.13 (p.120) in Toussaint 2013 (partial, work in progress)
    """
    tested_profile_order = (
        SHIKO,
        SON,
        SOUKOUS,
        RUMBA,
        GAHU,
        BOSSA,
    )  # sic GAHU, BOSSA

    # TODO Pressing, Lempel-Ziv, Entropy, "Metric", etc.

    # Distinct Distances
    assert [
        count_non_zero(indicator_to_interval(profile, adjacent_not_all=False))
        for profile in tested_profile_order
    ] == [4, 5, 7, 6, 7, 4]

    # nPVI
    assert [
        round(
            normalized_pairwise_variability_index(
                indicator_to_interval(profile)
            ),
            1,
        )
        for profile in tested_profile_order
    ] == [66.7, 40.5, 70.5, 41.0, 23.8, 14.3]


def test_toussaint_37_11():
    """
    Test the 16-unit rhythms with a beat pattern every 4th element
    against "Figure" 37.11 (p.290) in Toussaint 2013.
    TODO currently partial, work in progress, though note: several from the table covered elsewhere in the module:
    - 'Off-beatness'
    - Distinct durations
    """
    tested_profile_order = (
        SHIKO,
        SON,
        SOUKOUS,
        RUMBA,
        BOSSA,
        GAHU,
    )  # sic BOSSA, GAHU

    assert [is_maximally_even(profile) for profile in tested_profile_order] == [
        False,
        False,
        False,
        False,
        True,
        False,
    ]

    assert [
        has_oddity_property(profile) for profile in tested_profile_order
    ] == [
        False,
        True,
        False,
        True,
        True,
        False,
    ]

    # Distinct _adjacent_ distances. Note: _all_ distinct distances already tested above.
    assert [
        count_non_zero(
            indicator_to_interval(
                profile, adjacent_not_all=True, sequence_not_vector=False
            )
        )
        for profile in tested_profile_order
    ] == [2, 3, 4, 3, 2, 3]

    assert [has_deep_property(profile) for profile in tested_profile_order] == [
        True,
        False,
        False,
        False,
        True,
        False,
    ]


def test_wnbd_16():
    """
    Test the 16-unit rhythms with a beat pattern every 4th element
    against "Figure 8: Binary rhythms" in Gomez et al. (the WNBD paper)

    TODO attempt to resolve the divergences. Failing that:
    We include all of this information as part of our mission to make prior work accessible:
    readers navigating these papers should know about inconsistencies and errors.
    """
    for cycle, keith_value, off_beat_metric, wnbd in [
        (
            SHIKO,
            1,
            0,
            Fraction(6, 5),  # Confirm, 1.2
        ),
        (
            SON,
            2,
            1,
            Fraction(14, 5),  # Confirm, 2.8
        ),
        (
            RUMBA,
            2,
            2,
            Fraction(18, 5),  # Confirm, 3.6
        ),
        (
            SOUKOUS,
            3,
            2,
            Fraction(14, 5),  # TODO divergence. Paper has 3.6 (18/5)
        ),
        (
            GAHU,
            3,
            1,
            Fraction(16, 5),  # TODO divergence. Paper has 3.6 (18/5)
        ),
        (
            BOSSA,
            3,
            2,
            Fraction(16, 5),  # TODO divergence. Paper has 4.
        ),
    ]:

        assert keith_via_toussaint(cycle) == keith_value

        assert off_beat_metric == off_beatness(cycle)

        onset_beats = syncopation.vector_to_onset_beat(
            cycle, beat_unit_length=4
        )
        sm = syncopation.SyncopationMetric()
        assert (
            sm.weighted_note_to_beat_distance(onset_beats=onset_beats)
            == wnbd  # type: ignore
        )


def test_wnbd_12():
    """
    Test the 12-unit cycle with a beat pattern every 3rd element
    against the table "Figure 11: Ternary rhythms" in Gomez et al. (the WNBD paper)

    TODO attempt to resolve the divergences. Failing that:
    We include all of this information as part of our mission to make prior work accessible:
    readers navigating these papers should know about inconsistencies and errors.
    """
    for cycle, off_beat_metric, wnbd in [
        (
            SOLI,
            1,
            Fraction(
                12, 7
            ),  # TODO divergence. Paper has 2.142 (15/7, rounded down)
        ),
        (
            TAMBU,
            2,
            Fraction(
                12, 7
            ),  # TODO divergence. Paper has 2.142 (15/7, rounded down)
        ),
        (
            BEMBE,
            3,
            Fraction(18, 7),  # TODO divergence. Paper has 3 (21/7)
        ),
        (
            BEMBE_2,
            2,
            Fraction(
                12, 7
            ),  # TODO divergence. Paper has 2.142 (15/7, rounded down)
        ),
        (
            YORUBA,
            2,
            Fraction(18, 7),  # TODO divergence. Paper has 3 (21/7)
        ),
        (
            TONADA,
            1,
            Fraction(
                12, 7
            ),  # TODO divergence. Paper has 2.142 (15/7, rounded down)
        ),
        (
            ASAADUA,
            1,
            Fraction(
                12, 7
            ),  # TODO divergence. Paper has 2.142 (15/7, rounded down)
        ),
        (
            SORSONET,
            1,
            Fraction(18, 7),  # TODO divergence. Paper has 3 (21/7)
        ),
        (
            BEMBA,
            2,
            Fraction(
                12, 7
            ),  # TODO divergence. Paper has 2.142 (15/7, rounded down)
        ),
        (
            ASHANTI,
            2,
            Fraction(
                18, 7
            ),  # TODO divergence. Paper has 2 (14/7) *** only case of lower value. Others = 3/7 under
        ),
    ]:
        assert cycle.count(1) == 7  # Onsets
        assert len(cycle) == 12  # Grid

        assert off_beat_metric == off_beatness(cycle)

        onset_beats = vector_to_onset_beat(cycle, beat_unit_length=3)
        sm = syncopation.SyncopationMetric()
        assert (
            sm.weighted_note_to_beat_distance(onset_beats=onset_beats)
            == wnbd  # type: ignore
        )


def test_oddity_12():
    tested_profile_order = [
        SOLI,
        TAMBU,
        BEMBE,
        BEMBE_2,
        YORUBA,
        TONADA,
        ASAADUA,
        SORSONET,
        BEMBA,
        ASHANTI,
    ]
    assert [
        has_oddity_property(profile) for profile in tested_profile_order
    ] == [
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
    ]



import pytest

from amads.harmony.root_finding.parncutt import ParncuttRootAnalysis


def test_get_root_strength():
    """Test the get_root_strength method of ParncuttRootAnalysis."""
    analysis = ParncuttRootAnalysis([0, 4, 7], root_support_weights="v2")

    # Test with v2 weights
    assert (
        analysis.get_root_strength(0) == 10 + 3 + 5
    )  # Root + perfect fifth + major third
    assert analysis.get_root_strength(1) == 0  # No intervals from this pc
    assert analysis.get_root_strength(2) == 2 + 1  # Minor third + major second
    assert (
        analysis.get_root_strength(4) == 10
    )  # Root from perspective of this pc


def test_parn88_regression():
    """Test regression against known values from Parncutt (1988): Table 4"""

    # Helper function to test root ambiguity
    def test_ambiguity(expected, *chord, digits=1):
        analysis = ParncuttRootAnalysis(list(chord), root_support_weights="v1")
        assert round(analysis.root_ambiguity, digits) == expected

    # Dyads
    test_ambiguity(2.2, 0, 1)
    test_ambiguity(2.0, 0, 2)
    test_ambiguity(2.1, 0, 3)
    test_ambiguity(1.9, 0, 4)
    test_ambiguity(1.8, 0, 5)
    test_ambiguity(2.2, 0, 6)

    # Triads
    test_ambiguity(2.0, 0, 4, 7)
    test_ambiguity(2.1, 0, 3, 7)
    test_ambiguity(2.3, 0, 4, 8)
    test_ambiguity(2.5, 0, 3, 6)

    # Sevenths
    test_ambiguity(2.1, 0, 4, 7, 10)
    test_ambiguity(2.3, 0, 3, 7, 10)
    test_ambiguity(2.3, 0, 4, 7, 11)
    test_ambiguity(2.4, 0, 3, 6, 10)
    test_ambiguity(2.9, 0, 3, 6, 9)


def test_sanity_checks():
    """Test sanity checks for get_root finding and ambiguity"""
    # Test root finding
    analysis1 = ParncuttRootAnalysis([0, 4, 7])
    assert analysis1.root == 0

    analysis2 = ParncuttRootAnalysis([1, 4, 9])
    assert analysis2.root == 9

    # Test that diminished triad has higher ambiguity than major triad
    analysis3 = ParncuttRootAnalysis([0, 3, 6])
    analysis4 = ParncuttRootAnalysis([0, 4, 7])
    assert analysis3.root_ambiguity > analysis4.root_ambiguity


def test_root_support_versions():
    """Test different root support versions"""
    chord = [0, 4, 7, 10]  # Dominant seventh

    # Both should identify the same root
    analysis1 = ParncuttRootAnalysis(chord, root_support_weights="v1")
    analysis2 = ParncuttRootAnalysis(chord, root_support_weights="v2")
    assert analysis1.root == analysis2.root

    # But they should give different ambiguity values
    assert analysis1.root_ambiguity != analysis2.root_ambiguity


def test_available_root_support_weights():
    """Test the available root support weights"""
    # Test that both versions are available
    assert "v1" in ParncuttRootAnalysis.available_root_support_weights
    assert "v2" in ParncuttRootAnalysis.available_root_support_weights

    # Test that v1 has the expected values
    v1_weights = ParncuttRootAnalysis.available_root_support_weights["v1"]
    assert v1_weights[0] == 1.0
    assert v1_weights[7] == 1 / 2
    assert v1_weights[4] == 1 / 3

    # Test that v2 has the expected values
    v2_weights = ParncuttRootAnalysis.available_root_support_weights["v2"]
    assert v2_weights[0] == 10
    assert v2_weights[7] == 5
    assert v2_weights[4] == 3


def test_invalid_root_support_weights():
    """Test that invalid root support weights raise an error"""
    with pytest.raises(ValueError):
        ParncuttRootAnalysis([0, 4, 7], root_support_weights="invalid")


def test_empty_chord():
    """Test that an empty chord raises an error"""
    with pytest.raises(ValueError):
        ParncuttRootAnalysis([])



"""
Maximal correlation value's attribute and index pair from key_cc algorithm.

Corresponds to kkkey in miditoolbox.

Reference
---------
https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=68
"""

from itertools import chain
from typing import List, Optional, Tuple

from amads.core.basics import Score
from amads.pitch.key import profiles as prof
from amads.pitch.key.key_cc import key_cc

# TODO: include exceptions comments


def kkkey(
    score: Score,
    profile: prof.KeyProfile = prof.krumhansl_kessler,
    attribute_names: Optional[List[str]] = ["major", "minor"],
    salience_flag: bool = False,
) -> Tuple[str, int]:
    """
    Finds the pitch profile with the highest correlation value.

    Within `profile` there are multiple profiles named by attributes.
    This function returns the "best" attribute (string) and the best
    key (int) where the int corresponds to the 12 keys in order:
    0 -> C, 1 -> C#, ..., 11 -> B. (see key_cc.py for more details)

    Parameters
    ----------
    score: Score
        The musical score to analyze.
    profile: Profile
        The key profile to use for analysis.
    attribute_names: Optional[List[str]]
        List of attribute names that denote the particular PitchProfiles
        within the KeyProfile to compute correlations for.
        See key_cc for more details
    salience_flag: bool
        indicate whether we want to turn on salience weights in key_cc

    Returns
    -------
    tuple[str, int]
        The attribute name and key with the highest correlation coefficient.

    See Also
    --------
    key_cc
    """
    corrcoef_pairs = key_cc(score, profile, attribute_names, salience_flag)
    # list of pairs (attribute_name, [correlation coefficients])
    max_val_iter = (coefs for (_, coefs) in corrcoef_pairs if coefs is not None)
    # This code is a little unexpected: it first searches for the maximum
    # correlation value across all attributes and keys, then finds the
    # attribute and key index corresponding to that maximum value.
    max_val = max(chain.from_iterable(max_val_iter))
    nested_coefs_iter = (
        (attr, coefs.index(max_val))
        for (attr, coefs) in corrcoef_pairs
        if coefs is not None and max_val in coefs
    )
    return next(nested_coefs_iter)



"""
Determine if a musical score or its parts are monophonic.
"""

from amads.core.basics import Note, Part, Score


def _ismonophonic(notes: list[Note]):
    """
    Determine if a list of notes is monophonic.

    A monophonic list of notes has no overlapping notes (e.g., chords).
    Serves as a helper function for `ismonophonic` and
    `parts_are_monophonic`.

    Parameters
    ----------
    notes : list of Note
        The list of notes to analyze.

    Returns
    -------
    bool
        True if the list of notes is monophonic, False otherwise.
    """
    prev = None
    notes = list(notes)
    # Sort the notes by start time
    notes.sort(key=lambda note: note.onset)
    # Check for overlaps
    for note in notes:
        if prev:
            # 0.01 is to prevent precision errors when comparing floats
            if note.onset - prev.offset < -0.01:
                return False
        prev = note
    return True


def ismonophonic(score: Score):
    """
    Determine if a musical score is monophonic.

    A monophonic score has no overlapping notes (e.g., chords).

    Parameters
    ----------
    score : Score
        The musical score to analyze.

    Returns
    -------
    bool
        True if the score is monophonic, False otherwise.
    """
    return _ismonophonic(score.find_all(Note))


def parts_are_monophonic(score: Score) -> bool:
    """
    Determine if all parts of a musical score are monophonic.

    A monophonic part has no overlapping notes (e.g., chords).

    Parameters
    ----------
    score : Score
        The musical score to analyze.

    Returns
    -------
    bool
        True if all parts are monophonic, False otherwise.
    """
    for part in score.find_all(Part):
        if not _ismonophonic(part.find_all(Note)):
            return False
    return True



"""
Distribution of duration pairs in a Score.
Provides the `duration_distribution_2` function.

Can emulate the `ivdurdist2` function in Midi Toolbox.

Original doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=59
"""

from typing import Optional, Union, cast

from amads.core.basics import Note, Part, Score
from amads.core.distribution import Distribution
from amads.core.histogram import Histogram2D, centers_to_boundaries

_author_ = ["Yiming Huang", "Roger Dannenberg"]


def duration_distribution_2(
    score: Score,
    name: str = "Duration Pairs Distribution",
    bin_centers: Optional[list[float]] = None,
    ignore_extrema: Union[bool, str] = "unspecified",
    miditoolbox_compatible: bool = False,
) -> Distribution:
    r"""
    Returns the 2nd-order duration distribution of a musical score.

     Each duration is assigned to one of 9 bins.
    The default centers of the bins are on a logarithmic scale as follows:

    | component | bin center (in units of quarters) |
    |-----------|-----------------------------------|
    | 0         | 1/4 (sixteenth)                   |
    | 1         | sqrt(2)/4                         |
    | 2         | 1/2 (eighth)                      |
    | 3         | sqrt(2)/2                         |
    | 4         | 1 (quarter)                       |
    | 5         | sqrt(2)                           |
    | 6         | 2 (half)                          |
    | 7         | 2*sqrt(2)                         |
    | 8         | 4 (whole)                        |

    These centers can be overridden by providing a list of bin centers.

    If `midi_toolbox_compatible` is True, the behavior is the same as in
    the Midi Toolbox `durdist1` function, with each bin count increased by
    1e-12 to avoid division by zero, and values below $sqrt(2)/8$ quarters
    (just above a sixteenth triplet) and greater than $sqrt(2) \cdot 4$ quarters
    (about 5.65685) are ignored. Also, an error is raised if either
    `bin_centers` or `ignore_extrema` are specified.

    If `midi_toolbox_compatible` is False (default), the normal behavior
    is as follows. If `bin_centers` is not provided, the default centers
    listed above are used. Values below the lowest or above the highest
    bins are assigned to those bins appropriately.

    One remaining case is where you want to ignore values outside the
    range of the bins. This can be done by setting `ignore_extrema` to
    True. If `bin_centers` is not provided, the default bins are used,
    and the extreme boundaries are set to $sqrt(2)/8$ quarters and
    $sqrt(2) \cdot 4$ quarters respectively. If `bin_centers` is provided,
    they must include an additional lower bin and upper bin center value
    so that upper and lower boundaries can be computed as
    $sqrt(bin\\_centers[0] * bin\\_centers[1])$ and
    $sqrt(bin\\_centers[-2] * bin\\_centers[-1])$ respectively. Since values
    beyond the boundaries are ignored, the result will have
    $len(bin\\_centers) - 2$ bins.

    <small>**Authors**: Yiming Huang, Roger B. Dannenberg</small>

    Parameters
    ----------
    score : Score
        The musical score to analyze.
    name : str
        A name for the distribution and plot title.
    bin_centers : Union[list[float], None])
        bin centers (optional) (see description for details).
    ignore_extrema : bool
        If True, values outside the range of the bins are ignored
        (see description for details).
    miditoolbox_compatible : bool
        Matlab MIDI Toolbox avoids zero division by dividing counts
        by the total count plus 1e-12 times the number of counts.
        True enables this behavior as well as ignoring extreme values
        and setting the bins as described above. Default is False.
        If True, `bin_centers` and `ignore_extrema` must not be set.

    Returns
    -------
    Distribution
        containing and describing the distribution of note durations.

    Raises
    ------
    ValueError: If the score is not monophonic (e.g. contains chords) or
        if the miditoolbox is True and either `bin_centers` or
        `ignore_extrema` optional parameters are set or
        `ignore_extrema` is true and less than 3 bin centers are provided.
    """
    if miditoolbox_compatible:
        if bin_centers is not None or ignore_extrema != "unspecified":
            raise ValueError(
                "When miditoolbox_compatible is True, "
                "bin_centers and ignore_extrema must not be set."
            )
    if not score.ismonophonic():
        raise ValueError("Score must be monophonic")

    if ignore_extrema == "unspecified":
        ignore_extrema = miditoolbox_compatible
    else:
        if not isinstance(ignore_extrema, bool):
            raise ValueError(
                "ignore_extrema must be a boolean value or 'unspecified'"
            )
    initial_value = 1e-12 if miditoolbox_compatible else 0.0
    x_categories = None

    if not bin_centers:
        bin_boundaries = [2 ** ((i - 0.5) / 2.0) for i in range(-4, 6)]
        bin_centers = [2 ** (i / 2.0) for i in range(-4, 5)]
        ignore_extrema = True
        x_categories = [
            "sixteenth",
            "0.35",
            "eighth",
            "0.71",
            "quarter",
            "1.41",
            "half",
            "2.83",
            "whole",
        ]
    elif ignore_extrema:
        if len(bin_centers) < 3:
            raise ValueError(
                "When ignore_extrema is True and "
                "bin_centers is provided, at least "
                "three bin centers must be provided."
            )
        bin_boundaries = centers_to_boundaries(bin_centers, "log")
        bin_centers = bin_centers[1:-1]
    else:
        bin_boundaries = None

    if not x_categories:
        x_categories = [
            f"{bin_centers[i]:.2f}" for i in range(len(bin_centers))
        ]

    h = Histogram2D(
        bin_centers, bin_boundaries, "log", ignore_extrema, initial_value
    )  # type: ignore

    # TODO: what does matlab do with polyphony? This processes
    # each part separately (no interval from last note in one part
    # to the first note in the next part). This is probably
    # unnecessary since we already require monophonic input.

    # Ignore notes that are tied from previous notes by keeping track
    # of them in a set:
    tied_notes = set()
    for p in score.find_all(Part):
        part: Part = cast(Part, p)
        prev_dur = None
        prev_bin = None
        for n in part.find_all(Note):
            note: Note = cast(Note, n)
            if note in tied_notes:
                # skip tied notes
                continue
            if note.tie:
                # add tied notes to set so we can skip them later
                tied_notes.add(note)
            dur = note.duration
            prev_bin = h.add_point_2d(prev_dur, dur, 1.0, prev_bin)
            prev_dur = dur
            if prev_bin is None:
                prev_dur = None
    # normalize
    h.normalize()

    return Distribution(
        name,
        h.bins,
        "duration_pairs",
        [len(h.bins), len(h.bins)],
        x_categories,  # type: ignore
        "Duration (to)",
        x_categories,  # type: ignore
        "Duration (from)",
    )



import warnings
from math import isclose
from pathlib import Path
from typing import Optional, cast

from music21 import (
    chord,
    clef,
    instrument,
    key,
    layout,
    metadata,
    musicxml,
    note,
    stream,
    tempo,
    tie,
)
from music21.duration import Duration as m21Duration
from music21.meter.base import TimeSignature as m21TimeSignature

from amads.core.basics import (
    Chord,
    Clef,
    KeySignature,
    Measure,
    Note,
    Part,
    Rest,
    Score,
    Staff,
)
from amads.io.m21_show import music21_show

tied_notes = {}  # temporary data to track tied notes, this is a mapping
# from key number to Note object for notes that originate a tie. When we
# see a note that ends or continues a tie, we look up the origin of the
# tie in this dictionary and link it to the note that ends or continues
# the tie. Note that we do not encode 'let-ring' or 'continue-let-ring'
# in the AMADS data model, so we ignore those cases.
#     It is possible for notes to overlap in time and be tied. What ties
# to what in Music21 (as well as MIDI files) is ambiguous in this case,
# but ignoring overlapping ties would create multiple notes when there
# was only one in the MIDI file, so when there is overlap, we map from
# pitch to a list of notes that originate unterminated ties for this
# pitch. When a tie is terminated, we tie from the first note in the
# list. (First-in-first-out).


def _m21_clef(clef_str: str) -> clef.Clef:
    """translate AMADS clef ID string to music21 Clef object"""
    if clef_str == "treble":
        return clef.TrebleClef()  # type: ignore
    if clef_str == "bass":
        return clef.BassClef()  # type: ignore
    if clef_str == "alto":
        return clef.AltoClef()  # type: ignore
    if clef_str == "tenor":
        return clef.TenorClef()  # type: ignore
    if clef_str == "percussion":
        return clef.PercussionClef()  # type: ignore
    if clef_str == "treble8vb":
        return clef.Treble8vbClef()  # type: ignore
    raise ValueError(f"Unknown clef string: {clef_str}")


def fix_part_ids(musicxml_string):
    """Replace auto-generated part IDs with P1, P2, etc.

    The problem is that when I write PartStaff objects, Music21 generates
    unique IDs that are really ugly and not like other MusicXML files with,
    e.g., <part id="P1">, so this hack rewrites the XML with cleaner part
    names, as suggested by Claude after multiple attempts to set the ID's
    in the Music21 data structure.
    """
    import re

    # Find all unique part IDs
    part_ids = re.findall(
        r'<(?:score-)?part id="(P[a-f0-9]+)"', musicxml_string
    )
    unique_ids = list(dict.fromkeys(part_ids))  # Preserve order

    # Replace with simple IDs
    for i, old_id in enumerate(unique_ids, 1):
        musicxml_string = musicxml_string.replace(f'"{old_id}"', f'"P{i}"')

    return musicxml_string


def _add_measure_content(m21measure, measure, ties) -> None:
    """
    insert content (Notes, Chords, Rests) from a Measure to m21 measure

    This is complicated in that overlapping content must be separated
    out into voices.

    Parameters
    ----------
    m21measure: stream.measure
        The Music21 measure in which to add content

    measure: Measure
        The AMADS measure serving as a content source

    Returns
    -------
    float
        the maximum offset time of any

    """
    # see if there is any overlapping content
    need_voices = False
    for item, next_item in zip(measure.content, measure.content[1:]):
        if item.offset > next_item.onset + 1e-6:
            need_voices = True
            break
    if need_voices:  # collect voices - uses simple greedy algorithm
        # pick
        content = measure.content.copy()
        voices = []
        while len(content) > 0:
            # scan content, moving non-overlapping items into voice[].
            # overlapping items are saved at index i for the next voice(s)
            i = 0  # i is index of left-overs
            j = 0  # j is index to scan through content
            voice = []
            last_offset = 0
            for j in range(len(content)):
                if content[j].onset < last_offset - 1e-6:
                    content[i] = content[j]  # save for next voice(s)
                    i += 1
                else:
                    voice.append(content[j])
                    last_offset = content[j].offset
            # now, i is the length of overlapping items we want to keep
            del content[i:]
            voices.append(voice)

        for voice_num, voice in enumerate(voices):
            m21voice = stream.Voice(id=str(voice_num + 1))
            _add_measure_content_from_list(
                m21voice, measure, voice, measure.offset, ties
            )
            m21measure.insert(0, m21voice)
    else:
        _add_measure_content_from_list(
            m21measure, measure, measure.content, measure.offset, ties
        )


def _add_measure_content_from_list(m21parent, measure, content, dur, ties):
    """
    insert content (Notes, Chords, Rests) from a Measure to m21 stream

    The m21 stream, called `m21parent`, can be a Music21 measure or voice.
    Voices are needed when content overlaps. `content` must be non-overlapping.
    m21parent is padded with a rest to achieve a duration of `dur`
    """
    max_offset = measure.onset
    measure_position = 0  # keeps track of expected next onset time
    for item in content:
        measure_delta = item.onset - measure.onset
        duration = item.duration
        max_offset = max(max_offset, item.offset)
        if (
            isinstance(item, Note)
            or isinstance(item, Rest)
            or isinstance(item, Chord)
        ) and (measure_delta > measure_position + 1e-6):
            m21rest = note.Rest()
            m21rest.offset = measure_position
            m21rest.duration.quarterLength = measure_delta - measure_position
            m21parent.insert(measure_position, m21rest)

        # measure_position is the place where the item after this one
        # is expected. It is used to determine on the next iteration if
        # we need to pad with a rest
        measure_position = measure_delta + duration

        # now that we've padded to measure_position with a rest if necessary,
        # insert the Note, Rest, or Chord from the score:
        if isinstance(item, Note):
            m21note = note.Note(nameWithOctave=item.name_with_octave)
            m21note.offset = measure_delta
            m21note.duration.quarterLength = duration
            if isinstance(item.dynamic, int):
                m21note.volume.velocity = item.dynamic
            # otherwise, use default because I am not sure how
            #     to translate anything other dynamic value.
            m21parent.insert(m21note.offset, m21note)
            # if note is tied or tied to, enter it in ties:
            if item.tie:  # item is tied to item.tie
                is_tied_to = (item in ties) and ties[item][1]
                ties[item] = (m21note, is_tied_to)
                # set ties[item.tie]'s is_tied_to field to True
                tied_to_m21 = None
                if item.tie in ties:
                    tied_to_m21 = ties[item.tie][0]
                ties[item.tie] = (tied_to_m21, True)
            if (item in ties) and (ties[item][0] is None):
                # update entry with m21note
                ties[item] = (m21note, ties[item][1])
        elif isinstance(item, Rest):
            m21rest = note.Rest()
            m21rest.offset = measure_delta
            m21rest.duration.quarterLength = duration
            m21parent.insert(m21rest.offset, m21rest)
        elif isinstance(item, Chord):
            pitches = [
                n.name_with_octave for n in item.find_all(Note)  # type: ignore
            ]
            m21chord = chord.Chord(pitches)
            m21chord.offset = measure_delta
            m21chord.duration.quarterLength = duration
            m21parent.insert(m21chord.offset, m21chord)

    if max_offset < measure.offset - 1e-6:
        # need a rest to fill out the measure in Music21,
        # otherwise, it will shorten the duration of the measure
        m21rest = note.Rest()
        # music21 offset is relative to measure:
        m21rest.offset = max_offset - measure.onset
        m21rest.duration.quarterLength = measure.offset - max_offset
        m21parent.insert(m21rest.offset, m21rest)


def score_to_music21(
    score: Score, show: bool = False, filename: Optional[Path | str] = None
) -> stream.Score:
    """
    Convert a Score to music21

    Parameters
    ----------
    score: Score
        The Score to convert
    show : bool, optional
        If True, print the music21 score structure for debugging.
    filename : Optional[Path |  str]
        If `show` and not None, `filename` is shown

    Returns
    -------
    stream.Score
        The AMADS Score object converted to music21
    """
    # create a new music21 score
    m21score = stream.Score()
    m21score.metadata = metadata.Metadata()
    if score.has("title"):
        m21score.metadata.title = score.get("title")
    if score.has("composer"):
        m21score.metadata.composer = score.get("composer")

    # transfer tempo changes to music21
    for i, map_quarter in enumerate(score.time_map.changes):
        mm = score.time_map.get_tempo_at(i)
        m21score.insert(map_quarter.quarter, tempo.MetronomeMark(number=mm))

    ties = {}  # map from AMADS Note to (note.Note, tied_to)
    # where note.Note is the music21 note corresponding to the key or
    #     None if not processed yet, and tied_to is True if there is an
    #     incoming tie. (Outgoing ties can be detected by key.tie attribute)

    time_sigs = score.time_signatures
    for part_num, part in enumerate(score.find_all(Part)):
        part = cast(Part, part)
        part_id = f"P{part_num + 1}"
        staffs = part.list_all(Staff)
        make_staff_group = len(staffs) > 1
        m21staffs = []
        m21container = None  # make type checker happy
        for staff_num, staff in enumerate(staffs):
            if make_staff_group:
                id = f"P{part_num + 1}-Staff{staff_num + 1}"
                m21container = stream.PartStaff(id=id, partName=str(part_num))
                m21container.id = part_id
                m21container.groups.append(part_id)
                m21staffs.append(m21container)
            else:
                m21container = stream.Part(id=part_id)

            if part.instrument:
                try:
                    instr = instrument.fromString(part.instrument)
                except Exception as e:
                    warnings.warn(
                        f"Could not create instrument {part.instrument}: {e}"
                    )
                    instr = instrument.Instrument()
                    instr.instrumentName = part.instrument
                m21container.insert(0, instr)
            staff = cast(Staff, staff)
            time_sig_index = 0
            time_sig = time_sigs[time_sig_index]
            for index, measure in enumerate(staff.find_all(Measure)):
                measure = cast(Measure, measure)
                num = measure.number if measure.number else (index + 1)
                m21measure = stream.Measure(
                    number=num, duration=m21Duration(measure.duration)
                )
                # add time signature change if any
                if isclose(time_sig.time, measure.onset, abs_tol=1e-3):
                    if time_sig.upper != round(time_sig.upper):
                        raise ValueError(
                            "Cannot export fractional time"
                            f" signature {time_sig} to Music21."
                        )
                    ts_element = m21TimeSignature(
                        f"{round(time_sig.upper)}/{time_sig.lower}"
                    )
                    m21measure.insert(time_sig.time - measure.onset, ts_element)
                    # move to the next time_sig, if any left. If not, we do
                    # not change time_sig, but since subsequent measures have
                    # greater onset times, we won't try to add another time_sig
                    time_sig_index += 1
                    if time_sig_index < len(time_sigs):
                        time_sig = time_sigs[time_sig_index]
                # add key signature changes
                for ks in measure.find_all(KeySignature):
                    ks = cast(KeySignature, ks)
                    ks_element = key.KeySignature(ks.key_sig)
                    m21measure.insert(ks.onset - measure.onset, ks_element)
                # add clef changes
                for clef_change in measure.find_all(Clef):
                    clef_change = cast(Clef, clef_change)
                    try:
                        clef_element = _m21_clef(clef_change.clef)
                    except ValueError as e:
                        warnings.warn(str(e))
                        continue
                    m21measure.insert(
                        clef_change.onset - measure.onset, clef_element
                    )
                # add notes, rests, chords
                _add_measure_content(m21measure, measure, ties)
                m21container.append(m21measure)
            if time_sig_index != len(time_sigs):
                warnings.warn(
                    f"After converting AMADS staff {staff} to"
                    " Music21, there are left-over, unused time"
                    f" signatures, starting with {time_sig}."
                )
        # make a layout.StaffGroup to group the staffs if needed
        if make_staff_group:
            for m21staff in m21staffs:
                m21staff._idLastDeepCopyOf = part_id
                m21score.insert(0, m21staff)  # insert each staff
            staff_group = layout.StaffGroup(m21staffs, id=part_id)
            staff_group._idLastDeepCopyOf = part_id  # type: ignore
            m21score.insert(0, staff_group)  # this groups staffs into part
            # this makes no sense, but it works to give the "part" an id,
            # as suggested by Claude
            m21staffs[0].id = part_id
        else:
            m21score.insert(0, m21container)  # staff info is in an m21 Part

    # fix up ties
    for amads_note in ties.keys():
        # amads_note.tie points to the next note in the tie (or None if no
        # outgoing tie) is_tied_to indicates if there's an incoming tie
        # from a previous note
        (m21note, is_tied_to) = ties[amads_note]

        if m21note is None:  # note was tied to but we never put the tied-to
            # note into the m21score!
            warnings.warn(
                "In score_to_music21: tied-to note not found:"
                f" {amads_note.tie}"
            )
        elif amads_note.tie and not is_tied_to:
            # Note is tied to the next note, but not tied from a previous note
            m21note.tie = tie.Tie("start")
        elif amads_note.tie and is_tied_to:
            # Note is tied to the next note AND tied from a previous note
            m21note.tie = tie.Tie("continue")
        elif not amads_note.tie and is_tied_to:
            # Note is not tied to the next note, but is tied from a prev. note
            m21note.tie = tie.Tie("stop")
        elif not amads_note.tie and not is_tied_to:
            # No tie at all - don't set tie attribute
            pass
        else:
            assert (
                False
            ), f"Internal tie-fix-up error {amads_note}, {ties[amads_note]}"

    if show:
        music21_show(m21score, filename)  # type: ignore

    return m21score


def music21_export(
    score: Score,
    filename: Path | str,
    format: str,
    show: bool = False,
    display: bool = False,
) -> None:
    """Save a Score to a file in musicxml format using music21.

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    score : Score
        The Score to export.
    filename : Path | str
        The name or path of the file to save the MusicXML data.
    format : str
        The format to export. Must be "musicxml", "midi", or "kern".
    show : bool, optional
        If True, print the music21 score structure for debugging.
    display: bool, optional
        If True, open the generated PDF in the default viewer.
        Only relevant for PDF export (see `music21_xml_lilypond_export`).

    """
    m21score = score_to_music21(score, show, filename)

    if format == "musicxml":
        # Export as text and fix part id's
        exporter = musicxml.m21ToXml.GeneralObjectExporter(m21score)
        musicxml_bytes = exporter.parse()
        musicxml_str = musicxml_bytes.decode("utf-8")
        fixed_xml = fix_part_ids(musicxml_str)

        # Finally, write the file
        with open(filename, "w") as f:
            f.write(fixed_xml)
    elif format == "midi":
        m21score.write("midi", filename)
    elif format == "kern":
        m21score.write("kern", filename)



# fmt: off
# flake8: noqa E129,E303
"""
Pitch representation.
"""

import functools
from dataclasses import dataclass
from math import floor
from typing import Optional, Tuple, Union

from amads.core.vectors_sets import multiset_to_vector, weighted_to_indicator

CHROMATIC_NAMES = ["C", "C#", "D", "Eb", "E", "F",
                   "F#", "G", "Ab", "A", "Bb", "B"]
LETTER_TO_NUMBER = {"C": 0, "D": 2, "E": 4, "F": 5, "G": 7, "A": 9, "B": 11}
DIATONIC = [0, 2, 4, 5, 7, 9, 11]

__author__ = "Roger B. Dannenberg"

@functools.total_ordering
class Pitch:
    """Represents a symbolic musical pitch.

    A pitch is represented by a `key_num` and an `alt`. The `key_num`
    is a number that corresponds to the MIDI convention where C4 is 60, C# is
    61, etc., but generalized to floats (60.5 would be C4-quarter-tone-sharp).
    The `alt` is an alteration, where +1 represents a sharp and -1
    represents a flat. Alterations can also be, for example, 2 (double-sharp)
    or -0.5 (quarter-tone flat). The symbolic note name is derived by
    *subtracting* `alt` from `key_num`.

    E.g., C#4 has `key_num=61`, `alt=1`, so 61-1 gives us 60, corresponding
    to note name C. A Db has the same `key_num=61`, but alt=-1, and 61-(-1)
    gives us 62, corresponding to note name D. There is no representation
    for the “natural sign” (other than `alt=0`, which could imply no
    accidental) or “courtesy accidentals.”  Because accidentals normally
    “stick” within a measure or are implied by key signatures, accidentals
    are often omitted in the score presentation. Nonetheless, these
    implied accidentals are encoded in the `alt` attribute and `key_num`
    is the intended pitch with the accidental applied.

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    pitch : Union[int, float, str, None]
        Optional MIDI key_num or string Pitch name. Syntax is A-G followed
        by accidentals (see `accidental_chars` below) followed by octave
        number. (Defaults to 60)
    alt : Union[int, float, None], optional
        If pitch is a number, alt is an optional alteration (Defaults to 0).
        If `pitch - alt` does not result in a diatonic pitch number,
        alt is adjusted, normally choosing spellings C#, Eb, F#, Ab, and Bb.
        If pitch is a string, alt must be None.
    octave : Optional[int]
        If pitch is a string without an octave specification and `octave`
        is an int, then `octave` is used to specify the octave, where
        4 denotes the `key_num` range 60 through 71. `octave` defaults
        defaults to -1, which yields pitch class `key_num`s 0-11).
    accidental_chars: Union[list[str], None], optional
        Allows parsing of pitch names with customized accidental characters.
        The value is a tuple or list consisting of a string of flat characters
        and a string of sharp characters, e.g. `["fb", "s#"].
        (Defaults to None, which admits '♭', 'b' or '-' for flat, and
        '♯', '#', and '+' for sharp, but does not accept 'f' and 's'.)

    Attributes
    ----------
    key_num : float
        MIDI key number, e.g., C4 = 60, generalized to float.
    alt : float
        Alteration, e.g., flat = -1.

    Examples
    --------
    >>> p = Pitch(64)
    >>> p
    Pitch(name='E4', key_num=64)

    >>> p.octave
    4

    >>> p = Pitch("E4")
    >>> p.octave
    4

    >>> p = Pitch("F#################2")
    >>> p.alt
    17

    >>> p.octave
    2

    >>> p = Pitch("E--------------------4")
    >>> p.alt
    -20

    >>> p.octave
    4

    >>> p.register
    2

    >>> Pitch(61.5, alt=1.5)
    Pitch(name='C?4', key_num=61.5)

    >>> # key_num - alt must be a diatonic pitch number. If not, key_num
    >>> # gets priority and alt is adjusted to the smallest valid value.
    >>> # Here, alt is adjusted to 0, which preserves key_num 60:
    >>> Pitch(60, alt=1.4)
    Pitch(name='C4', key_num=60)

    """
    __slots__ = ["key_num", "alt"]
    key_num: float  # but None is also allowed
    alt: float

    def _fix_alteration(self) -> None:
        """Fix the alteration to ensure it is a valid value.
        
        I.e., that `(key_num - alt) % 12` denotes one of {C D E F G A B}.
        If the invariant is false, we give priority to key_num and find
        the smallest value of alt that makes it true. For integer key_num,
        we break ties (enharmonics) by favoring C#, Eb, F#, Ab, and Bb.
        """
        unaltered = self.key_num - self.alt
        # if alt is off by a tiny amount, we just correct it
        if abs(unaltered - round(unaltered)) < 1e-6:
            unaltered = round(unaltered)
            self.alt = self.key_num - unaltered
        # if unaltered was even close to an integer, it will become an
        # integer so the following test is correct:
        if isinstance(unaltered, int) and (unaltered % 12) in DIATONIC:
            return  # valid key_num and alt

        # If alt is not an integer, we adjust it to be as small as
        # possible (< 1 when unaltered is in C-to-E or G-to-B, and < 0.5
        # when pc is in E-to-F or B to C).  First, we need to force
        # key_num to be an integer. For 3.10 compatibility, can only
        # send .is_integer() to a float:
        if not float(self.key_num).is_integer():
            # if alt could be less than 0.5, make it so:
            closest_pc = round(self.key_num) % 12
            self.alt = self.key_num - round(self.key_num)
            # now alt < 0.5
            if not closest_pc in DIATONIC:
                sign = 1 if self.alt > 0 else -1
                self.alt -= sign
            assert abs(self.alt) < 1, "alt must be < 1 in magnitude"
            # now we have a small value for a non-integer alt
        else:
            # again, we have to change alt (an integer this time). Make it 
            # -1, 0, or 1, giving priority to default spellings C#, Eb, F#,
            # Ab, and Bb.
            pc = self.key_num % 12
            if pc in [1, 6]:  # C#->C, F#->F
                self.alt = 1
            elif pc in [3, 8, 10]:  # Eb->E, Ab->A, Bb->B
                self.alt = -1
            else:
                self.alt = 0
        unaltered = self.key_num - self.alt
        assert (unaltered - round(unaltered)) < 1e-6
        assert round(unaltered) % 12 in DIATONIC, "pc must be in valid"


    def __init__(self,
                 pitch: Union["Pitch", int, float, str, None] = 60,
                 alt: Union[int, float, None] = None,
                 octave: Union[int, None] = None,
                 accidental_chars: Optional[str] = None):
        if isinstance(pitch, str):
            if alt is not None:
                raise ValueError("If pitch is a string, alt must be None")
            self.key_num, self.alt = Pitch.from_name(pitch, octave, 
                                                     accidental_chars)
        elif isinstance(pitch, Pitch):
            self.key_num = pitch.key_num
            self.alt = pitch.alt
        elif pitch is None:
            self.key_num = None  # type: ignore (None is allowed, but if we put
            # that in the type annotation, we have to annotate every use of
            # arithmetic on key_num.)
            self.alt = (0 if alt is None else alt)
        else:  # pitch is a number (int or float)
            # this will raise a ValueError if pitch is not some kind of number:
            pitch = float(pitch)  # converts numpy.int64, nympy.floating, etc.
            if pitch.is_integer():  # for nicer printing
                pitch = int(pitch)  # pitch numbers as integers.
            self.key_num = pitch
            self.alt = (0 if alt is None else alt)
            self._fix_alteration()


    def __repr__(self):
        return f"Pitch(name='{self.name_with_octave}', key_num={self.key_num})"

    
    def as_tuple(self):
        """Return a tuple representation of the `Pitch` instance.

        Returns
        -------
        tuple
            A tuple containing the `key_num` and `alt` values.
        """
        return (self.key_num, self.alt)


    def __eq__(self, other):
        """Check equality of other to a Pitch instance.

        Pitches are equal if they are both Pitches, and
        both `key_num` and `alt` are equal. Enharmonics are therefore
        not equal, but enharmonic equivalence can be written simply as
        `p1.key_num == p2.key_num`

        Parameters
        ----------
        other : Pitch
            The other Pitch instance to compare with.

        Returns
        -------
        bool
            True if the `key_num` and `alt` values are equal, False otherwise.
        """
        return isinstance(other, Pitch) and self.as_tuple() == other.as_tuple()


    def __hash__(self) -> int:
        """Return a hash value for the Pitch instance.

        Returns
        -------
        int
            A hash value representing the Pitch instance.
        """
        return hash(self.as_tuple())


    def __lt__(self, other) -> bool:
        """Check if this Pitch instance is less than another Pitch instance.
        Pitches are compared first by `key_num` and then by `alt`. Pitches
        with sharps (i.e. positive alt) are considered lower because
        their letter names are lower in the musical alphabet.

        Parameters
        ----------
        other : Pitch
            The other Pitch instance to compare with.

        Returns
        -------
        bool
            True if this Pitch instance is less than the other, False otherwise.
        """
        return (self.key_num, -self.alt) < (other.key_num, -other.alt)


    @classmethod
    def from_name(cls, name: str,
                  octave: Optional[float] = -1,
                  accidental_chars: Optional[str] = None
                  ) -> Tuple[float, float]:
        """
        Converts a string like "Bb" to a (pitch, alt) tuple.

        For example, converts "Bb" to (10, -1).
        If the string has an octave number or octave is given, the
        octave will be applied, e.g., "C4" yields (60, 0). `octave` takes
        effect if it is a number and name does not include an octave.
        If both `name` has no octave and `octave` is None, the octave is
        -1, yielding pitch class numbers 0-11.

        The first character must be one of the unmodified base pitch names:
        C, D, E, F, G, A, B (not case-sensitive).

        Subsequent characters must indicate a single accidental type:
        one of '♭', 'b' or '-' for flat, and '♯', '#', and '+' for sharp,
        unless accidental_chars specified exactly the acceptable flat
        and sharp chars, e.g., "fs" indicates 'f' for flat, 's' for sharp.

        Note that 's' is not a default accidental type as it is ambiguous:
        'Fs' probably indicates F#, but Es is more likely Eb (German).

        Also unsupported are:
        mixtures of sharps and flats (e.g., B#b);
        symbols for double sharps, quarter sharps, naturals, etc.;
        any other characters (except space, tab and underscore,
        which are allowed but ignored).

        Following accidentals (if any) is an optional single-digit octave
        number. Note that MIDI goes below C0; if the octave number is
        omitted, the octave will be -1, which corresponds to pitch class
        numbers 0-11 and octave -1 (which you can also specify as octave).

        Instructive error messages are given for invalid input.

        Parameters
        ----------
        name : str
            The string representation of the pitch name.
        octave : Optional[float], optional
            The octave number if not specified in the name.
            (Defaults to -1)
        accidental_chars : Optional[str], optional
            The characters to use for flat and sharp accidentals.
            (Defaults to None, which admits '♭', 'b' or '-' for flat, and
            '♯', '#', and '+' for sharp.)

        Returns
        -------
        Tuple[float, float]
            A tuple containing the `key_num` and `alt` values.
        """
        name = name.replace(" ", "").replace("\t", "").replace("_", "")
        if name == "":
            return 60, 0
        pitch_base = name[0].upper()  # error if non-string
        if pitch_base not in "ABCDEFG":
            raise ValueError("Invalid first character: must be one of ABCDEFG")
        pitch_class = LETTER_TO_NUMBER[pitch_base]

        name = name[1:]  # trim the note letter
        if len(name) > 0:
            if name[-1].isdigit():  # final character indicates octave
                octave = int(name[-1])  # overrides octave parameter
                name = name[:-1]  # remove octave from working
        if octave is None:  # no octave given in name or 2nd parameter
            octave = -1

        # parse the accidentals, if any
        if accidental_chars:
            flat_chars = [accidental_chars[0]]
            sharp_chars = [accidental_chars[1]]
        else:
            flat_chars = ["♭", "b", "-"]
            sharp_chars = ["♯", "#", "+"]
        if all(x in flat_chars for x in name):  # flats
            alteration = -len(name)
        elif all(x in sharp_chars for x in name):  # sharps
            alteration = len(name)
        else:
            raise ValueError("Invalid accidentals: must be only " +
                             f"{flat_chars} or {sharp_chars}.")

        # note that octave applies to pitch_class before alteration, so B#3=C4
        return pitch_class + 12 * (octave + 1) + alteration, alteration


    def get_name(self, accidental_chars: str = "b#") -> str:
        """Return string name including accidentals (# or b) but no octave.

        See the `name` property for details.

        Parameters
        ----------
        accidental_chars : str, optional
            The characters to use for flat and sharp accidentals.
            (Defaults to "b#")

        Returns
        -------
        str
            The string representation of the pitch name, including accidentals.
        """
        accidentals = "?"
        sharp_char = (accidental_chars[1] if len(accidental_chars) > 1 else "#")
        if round(self.alt) == self.alt:  # an integer value
            if self.alt > 0:
                accidentals = sharp_char * round(self.alt)
            elif self.alt < 0:
                accidentals = accidental_chars[0] * round(-self.alt)
            else:
                accidentals = ""  # natural
        return self.step + accidentals


    def get_name_with_octave(self, accidental_chars: str = "b#") -> str:
        """Return string name with octave, e.g., C4, B#3, etc.

        See the [name_with_octave][amads.core.pitch.Pitch.name_with_octave]
        property for details.

        Parameters
        ----------
        accidental_chars : str, optional
            The characters to use for flat and sharp accidentals.
            (Defaults to "b#")

        Returns
        -------
        str
            The string representation of the pitch name with octave.
        """
        return ("unpitched" if self.key_num is None
                            else self.name + str(self.octave))


    @property
    def step(self) -> str:
        """The diatonic name of the pitch: A, B, C, D, E, F, or G.

        The diatonic name corresponds to letter name without accidentals.

        Returns
        -------
        str
            The name of the pitch, a letter in "A" through "G".
        """
        if self.key_num is None:
            return "unpitched"
        unaltered = round(self.key_num - self.alt)
        return ["C", "?", "D", "?", "E", "F", "?", "G", "?", "A", "?", "B"][
            unaltered % 12]


    @property
    def name(self) -> str:
        """The string name including accidentals (# or b).

        The octave number is omitted. If alt is not an integer,
        return the step name concatenated with "?". See also
        [get_name][amads.core.pitch.Pitch.get_name], which
        accepts a parameter to specify accidental characters.
        """
        return self.get_name()


    @property
    def name_with_octave(self) -> str:
        """The string name with octave, e.g., "C4", "B#3", etc.

        The octave number is calculated by

            (key_num - alteration) // 12 + 1  # (integer division)

        and refers to the pitch before alteration, e.g., C4 is
        enharmonic to B#3 and represents the same (more or less)
        pitch even though the written octave numbers differ.

        See also [get_name_with_octave]
        [amads.core.pitch.Pitch.get_name_with_octave],
        which accepts a parameter to specify custom characters
        to represent accidentals.
        """
        return self.get_name_with_octave()


    @property
    def octave(self) -> int:
        """The octave number of the note name.

        The note name is based on `key_num - alt`, e.g.,
        C4 has octave 4 while B#3 has octave 3.
        """
        unaltered = round(self.key_num - self.alt)
        return (unaltered // 12) - 1


#    @octave.setter
#    def octave(self, oct: int) -> None:
#        """Set the octave number of the note.
#
#        Parameters
#        ----------
#        oct : int
#            The new octave number.
#        """
#        old_oct = self.octave
#        self.key_num += (oct - old_oct) * 12


    @property
    def pitch_class(self) -> int:
        """The pitch class of the note, e.g., 0, 1, 2, ..., 11.

        The pitch class is the `key_num modulo 12`, which gives the
        class of this pitch in the range 0-11. If the `key_num` is
        non-integer, it is rounded.

        Returns
        -------
        int
            The pitch class of the note.
        """
        return round(self.key_num) % 12


# setters seem dangerous since Pitch is nominally immutable and
# often shared across multiple Note objects.
#
#    @pitch_class.setter
#    def pitch_class(self, pc: int) -> None:
#        """Set the pitch class of the note. The resulting
#        `register` will be the same. E.g., for B#3
#        (`key_num == 60`, `register == 4`), setting the
#        `pitch_class` to 2 will yield `key_num == 62`, 
#        `register == 4`, although the `octave` changes from
#        3 to 4. Setting the pitch class to 0 will result
#        in a pitch name of "C4".  The resulting `alt`
#        (accidental) values result in names of 
#        C, C#, D, Eb, E, F, F#, G, Ab, A, Bb, and B.
#
#        Parameters
#        ----------
#        pc : int
#            The new pitch class value.
#        """
#        self.key_num = (self.octave + 1) * 12 + pc % 12
#        self.alt = 0
#        self._fix_alteration()


    @property
    def register(self) -> int:
        """Returns the absolute octave number based on `floor(key_num)`.
        
        Both C4 and B#3 have register 4.
        """
        return floor(self.key_num) // 12 - 1


#    @register.setter
#    def register(self, reg: int) -> None:
#        """Set the register of the note.
#
#        Parameters
#        ----------
#        reg : int
#            The new register
#        """
#        old_reg = self.register
#        self.key_num += (reg - old_reg) * 12

    def enharmonic(self) -> "Pitch":
        """Construct an enharmonic equivalent.

        If `alt` is non-zero, return a Pitch where `alt` is zero
        or has the opposite sign and where `alt` is minimized. E.g.
        enharmonic(Cbb) is A# (not Bb). If alt is zero, return a
        Pitch with alt of +1 or -1 if possible. Otherwise, return
        a Pitch with alt of -2 (Ebb, Abb or Bbb).
        Note the difference between this and `simplest_enharmonic`.

        Returns
        -------
        Pitch
            A new Pitch object representing the enharmonic equivalent.

        Examples
        --------
        >>> Pitch("C4").enharmonic()
        Pitch(name='B#3', key_num=60)

        >>> Pitch("B3").enharmonic()
        Pitch(name='Cb4', key_num=59)

        >>> Pitch("B#3").enharmonic()
        Pitch(name='C4', key_num=60)

        >>> bds = Pitch("B##3")
        >>> bds.enharmonic() # change of direction
        Pitch(name='Db4', key_num=61)

        >>> bds.upper_enharmonic()  # note the difference
        Pitch(name='C#4', key_num=61)

        >>> Pitch("Dbb4").enharmonic()
        Pitch(name='C4', key_num=60)
        """
        alt = self.alt
        unaltered = round(self.key_num - alt)
        if alt < 0:
            while alt < 0 or (unaltered % 12) not in [0, 2, 4, 5, 7, 9, 11]:
                unaltered -= 1
                alt += 1
        elif alt > 0:
            while alt > 0 or (unaltered % 12) not in [0, 2, 4, 5, 7, 9, 11]:
                unaltered += 1
                alt -= 1
        else:  # alt == 0
            unaltered = unaltered % 12
            if unaltered in [0, 5]:  # C->B#, F->E#
                alt = 1
            elif unaltered in [11, 4]:  # B->Cb, E->Fb
                alt = -1
            else:  # A->Bbb, D->Ebb, G->Abb
                alt = -2
        return Pitch(self.key_num, alt)


    def simplest_enharmonic(self,
            sharp_or_flat: Optional[str] = 'default') -> "Pitch":
        """
        Create Pitch object with the simplest enharmonic representation.

        I.e., if there exists an enharmonic-equivalent pitch with no
        alterations, then use that. If the Pitch is already in simplest
        form (e.g., C4), it is simply returned. If an alteration is
        needed, then use sharps or flats depending on `sharp_or_flat`.
        If `sharp_or_flat` is omitted, the same enharmonic choice
        as the Pitch constructor is used (C#, Eb, F#, Ab, and Bb).

        Parameters
        ----------
        sharp_or_flat: str
            This is only relevant if the pitch needs an alteration, otherwise
            it is unused. The value can be "sharp" (use sharps), "flat" (use
            flats), and otherwise use the same enharmonic choice as the Pitch
            constructor.

        Examples
        --------

        >>> bds = Pitch("B##3")
        >>> bds.simplest_enharmonic()
        Pitch(name='C#4', key_num=61)

        >>> bds.simplest_enharmonic(sharp_or_flat="flat")
        Pitch(name='Db4', key_num=61)

        >>> Pitch("C4").simplest_enharmonic()
        Pitch(name='C4', key_num=60)

        Returns
        -------
        Pitch
            A Pitch object representing the enharmonic equivalent.
        """
        if self.alt in [None, 0]:
            return self

        if self.pitch_class in [0, 2, 4, 5, 7, 9, 11]:  # C, D, E, F, G, A, B
            return Pitch(self.key_num)
        elif sharp_or_flat == "sharp":  # unaltered in 1, 3, 6, 8, 10
            return Pitch(self.key_num, 1)
        elif sharp_or_flat == "flat":
            return Pitch(self.key_num, -1)
        else:  # let Pitch figure out which enharmonic spelling (alt) to use:
            return Pitch(self.key_num)


    def upper_enharmonic(self) -> "Pitch":
        """
        Return the enharmonic based on the note name above.

        The result will have the next higher diatonic name
        with `alt` accordingly decreased by 1 or 2, e.g.,
        C#->Db, C##->D, Cb->Dbbb.

        Returns
        -------
        Pitch
            A Pitch object representing the upper enharmonic equivalent.


        Examples
        --------
        >>> bds = Pitch("B##3")
        >>> bds
        Pitch(name='B##3', key_num=61)

        >>> cis = bds.upper_enharmonic()
        >>> cis
        Pitch(name='C#4', key_num=61)

        >>> des = cis.upper_enharmonic()
        >>> des
        Pitch(name='Db4', key_num=61)

        >>> des.upper_enharmonic()
        Pitch(name='Ebbb4', key_num=61)

        >>> Pitch("D4").upper_enharmonic()
        Pitch(name='Ebb4', key_num=62)

        """
        alt = self.alt
        unaltered = round(self.key_num - alt) % 12
        if unaltered in [0, 2, 5, 7, 9]:  # C->D, D->E, F->G, G->A, A->B
            alt -= 2
        else:  # E->F, B->C
            alt -= 1
        return Pitch(self.key_num, alt)


    def lower_enharmonic(self) -> "Pitch":
        """Return the enharmonic based on the note name below.

        The result will have the next lower diatonic name
        with `alt` accordingly increased by 1 or 2, e.g.,
        Db->C#, D->C##, D#->C###.

        Returns
        -------
        Pitch
            A Pitch object representing the lower enharmonic equivalent.

        Examples
        --------
        >>> Pitch("Db4").lower_enharmonic()
        Pitch(name='C#4', key_num=61)

        >>> Pitch("D4").lower_enharmonic()
        Pitch(name='C##4', key_num=62)

        >>> Pitch("C#4").lower_enharmonic()
        Pitch(name='B##3', key_num=61)

        """
        alt = self.alt
        unaltered = round(self.key_num - alt) % 12
        if unaltered in [2, 4, 7, 9, 11]:  # D->C, E->D, G->F, A->G, B->A
            alt += 2
        else:  # F->E, C->B
            alt += 1
        return Pitch(self.key_num, alt)


@dataclass
class PitchCollection:
    """
    Combined representations of more than one pitch. Differs from Chord
    which has onset, duration, and contains Notes, not Pitches.

    Parameters
    ----------
    pitches : list[Pitch]
        A list of Pitch instances.

    Attributes
    ----------
    pitches : list[Pitch]
        A list of Pitch instances.

    Examples
    --------
    >>> test_case = ['G#4', 'G#4', 'B4', 'D4', 'F4', 'Ab4']  
    >>> pitches = [Pitch(p) for p in test_case]  
    >>> pitches_gathered = PitchCollection(pitches)

    >>> pitches_gathered.pitch_name_multiset  
    ['G#4', 'G#4', 'B4', 'D4', 'F4', 'Ab4']

    >>> pitches_gathered.pitch_num_multiset  
    [68, 68, 71, 62, 65, 68]

    >>> pitches_gathered.pitch_class_multiset  
    [2, 5, 8, 8, 8, 11]

    >>> pitches_gathered.pitch_class_set  
    [2, 5, 8, 11]

    >>> pitches_gathered.pitch_class_vector  
    (0, 0, 1, 0, 0, 1, 0, 0, 3, 0, 0, 1)

    >>> pitches_gathered.pitch_class_indicator_vector
    (0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1)
    """
    __slots__ = ["pitches"]
    pitches: list[Pitch]

    @property
    def pitch_num_multiset(self):
        """
        Return a list of pitch numbers from the pitches in the collection.        
        """
        return [p.key_num for p in self.pitches]

    @property
    def pitch_name_multiset(self):
        """
        Return a list of pitch names with octaves from the pitches in the collection.        
        """
        return [p.name_with_octave for p in self.pitches]

    @property
    def pitch_class_multiset(self):
        """
        Return a sorted list of pitch classes from the pitches in the collection, including duplicates.        
        """
        return sorted([p.pitch_class for p in self.pitches])

    @property
    def pitch_class_set(self):
        """ 
        Return a sorted list of pitch classes from the pitches in the collection without duplicates.      
        """
        return sorted(list(set(self.pitch_class_multiset)))

    @property
    def pitch_class_vector(self):
        """
        Return a pitch class vector (12-dimensional) representing the count of each pitch class in the collection.
        """
        return multiset_to_vector(self.pitch_class_multiset, max_index=11)

    @property
    def pitch_class_indicator_vector(self):
        """
        Return a pitch class indicator vector (12-dimensional) representing the presence (1) or absence (0) of each pitch class in the collection.
        """
        return weighted_to_indicator(self.pitch_class_vector)



"""
The Distribution class represents distributions and distribution metadata.
"""

from typing import Any, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from amads.algorithms.norm import normalize

# We should not force this on users as it is not compatible with all backends:
# matplotlib.use('TkAgg')

__author__ = ["Roger Dannenberg", "Tai Nakamura", "Di Wang"]


# def norm_1d(
#     profile: Iterable,
#     round_places: Optional[int] = None,
# ) -> np.ndarray:
#     """
#     Copied implementation from algorithms/norm.py with some modification
#     because using the original will cause a circular dependency in the package

#     Notes
#     -----
#     - This normalizes a 1-D vector to L1 (sum to 1). If `round_places` is
#       provided, the result is rounded for display/consistency.
#     """
#     norm_ord = 1
#     profile_array = np.asarray(profile)
#     norm_dist = profile_array / np.linalg.norm(profile_array, ord=norm_ord)

#     if round_places is None:
#         return norm_dist
#     elif isinstance(round_places, int):
#         return np.round(norm_dist, round_places)
#     else:
#         raise ValueError(
#             f"invalid round_places parameter {round_places}, expected int or None"
#         )


class Distribution:
    """
    Represents a probability distribution or histogram and its metadata.

    See histogram.py for Histogram1D and Histogram2D, which facilitate the
    computation of distributions from data.

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    name: str
        The name of the distribution used for plot titles.

    data: List[Any]
        The data points for the distribution.

    distribution_type: str
        The type of distribution. Currently used values are described
        below. "weights" can mean either probabilities or raw counts.

        - "pitch_class" - weights for 12 pitch classes, possibly weighted
            by duration.
        - "interval" - weights for pitch intervals, possibly weighted by
            duration.
        - "pitch_class_interval" - weights for pitch class intervals,
            possibly weighted by duration.
            (mod 12, not currently used; should there also be a
            pitch_class_size based on absolute interval size mod 12?
            And if so, should there be an interval_class computed from
            interval mod 12? And interval_size_class based on absolute
            interval size mod 12? Note that "interval" ignores intervals
            larger than one octave.)
        - "duration" - weights for durations
        - "interval_size" - weights for interval sizes, possibly weighted
            by duration.
        - "interval_direction" - proportion of upward intervals for each
            interval size, possibly weighted by duration.
        - "pitch_class_transition" - weights for pitch class transitions,
            possibly weighted by duration (2-dimensional distribution).
        - "interval_transition" - weights for interval transitions,
            possibly weighted by duration (2-dimensional distribution).
        - "duration_transition" - weights for duration transitions,
            possibly weighted by duration (2-dimensional distribution).
        - "key_correlation" - weights for key correlations, generally
            correlation with 12 major key profiles followed by
            correlations with 12 minor key profiles.
        - "symmetric_key_profile" - weights for symmetric key profiles.
            Key profiles themselves are distributions. Symmetric keys
            use the same 12 weights for all keys (e.g., Krumhansl-Kessler),
            simply rotated for each key.
        - "asymmetric_key_profile" - weights for asymmetric key profiles,
            where each key has its own set of 12 weights (e.g., Bellman-Budge).
        - "root_support_weights" - root support weights (see
            `amads.harmony.root_finding.parncutt`)

        This list is open-ended and is currently just informational.
        The value is not used for plotting or any other purpose.

    dimensions : List[int]
        The dimensions of the distribution, e.g.
        [12] for a pitch class distribution or [25, 25] for an
        interval_transition (intervals are from -12 to +12 and include
        0 for unison, intervals larger than one octave are ignored).

    x_categories: List[Union[int, float, str]]
        The categories for the x-axis.

    x_label: str
        The label for the x-axis.

    y_categories: Optional[List[Union[int, float, str]]]
        The categories for the y-axis, if any, otherwise None.

    y_label: str
        The label for the y-axis.

    Attributes
    ----------
    same as Parameters (above)
    """

    # Class variable documenting allowable 1-D plotting styles
    POSSIBLE_1D_PLOT_OPTIONS = ["bar", "line"]
    # Default color for 1-D bar charts
    DEFAULT_BAR_COLOR = "skyblue"

    def __init__(
        self,
        name: str,
        data: List[Any],
        distribution_type: str,
        dimensions: List[int],
        x_categories: List[Union[int, float, str]],
        x_label: str,
        y_categories: Optional[List[Union[int, float, str]]],
        y_label: str,
    ):
        """
        Initialize a Distribution instance.
        """
        self.name = name
        self.data = data
        self.distribution_type = distribution_type
        self.dimensions = dimensions
        self.x_categories = x_categories
        self.x_label = x_label
        self.y_categories = y_categories
        self.y_label = y_label

    def normalize(self):
        """
        Convert weights or counts to a probability distribution that sums to 1.
        """
        self.data = normalize(self.data, "Sum").tolist()
        return self

    def plot(
        self,
        color: Optional[str] = None,
        option: Optional[str] = None,
        show: bool = True,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
    ) -> Figure:
        """
        Virtual plot function for Distribution.
        Allows standalone plotting of a Distribution (when fig and ax are None),
        while providing enough extensibility to invoke this plot function or its
        overwritten variants for subplotting when fig and ax are provided as
        arguments.

        Parameters
        ----------
        color : Optional[str]
            Plot color string specification. In this particular plot function,
            it is handled in 1-D distributions and ignored in 2-D distributions.
            None for default option (Distribution.DEFAULT_BAR_COLOR).
        option : Optional[str]
            Plot style string specification. In this particular plot function,
            only {"bar", "line"} are valid string arguments that will be handled
            in a 1-D distribution, while any argument is ignored in 2-D
            distributions. None for default option ("bar").
        show : bool
            Whether to call ``plt.show()`` at the end.
        fig : Figure
            Provide existing Figure to draw on; if omitted, a new
            figure is created.
        ax : Axes
            Provide existing axes to draw on; if omitted, a new
            figure and axes are created.

        Raises
        ------
        ValueError
            A ValueError is raised if:

            - `ax` (axes) but not `fig` (Figure) is provided
            - `dims` is not 1 or 2

        Notes
        -----
        Behavior to this specific plot method:

        - 1-D: bar (default) or line when kind is "line"
        - 2-D: heatmap
        """
        dims = len(self.dimensions)
        if dims not in (1, 2):
            raise ValueError(
                "Unsupported number of dimensions for Distribution class"
            )

        # Figure/axes handling: either both `fig` and `ax` are provided, or
        # neither; in the latter case, create a new figure/axes pair.
        if fig is None:
            if ax is not None:
                raise ValueError("invalid figure/axis combination")
            fig, ax = plt.subplots()
        else:
            if ax is None:
                raise ValueError("invalid figure/axis combination")

        if dims == 1:
            if color is None:
                color = Distribution.DEFAULT_BAR_COLOR
            if option is None:
                option = "bar"
            x = range(len(self.x_categories))
            # 1-D distributions: draw either a bar chart or a line chart.
            if option == "bar":
                ax.bar(x, self.data, color=color)
            elif option == "line":
                ax.plot(x, self.data, color=color, marker="o")
            else:
                raise ValueError(f"unknown kind for 1D plot: {option}")

            ax.set_xticks(list(x))
            ax.set_xticklabels([str(label) for label in self.x_categories])
            ax.set_xlabel(self.x_label)
            ax.set_ylabel(self.y_label)
            ax.set_title(self.name)

        else:  # dims == 2
            # 2-D distributions: render as a heatmap with a colorbar.
            data = np.asarray(self.data)
            cax = ax.imshow(
                data, cmap="gray_r", aspect="auto", interpolation="nearest"
            )
            fig.colorbar(cax, ax=ax, label="Proportion")

            ax.set_xlabel(self.x_label)
            ax.set_ylabel(self.y_label)
            ax.set_title(self.name)

            ax.set_xticks(range(len(self.x_categories)))
            ax.set_xticklabels([str(label) for label in self.x_categories])
            if self.y_categories is not None:
                ax.set_yticks(range(len(self.y_categories)))
                ax.set_yticklabels([str(label) for label in self.y_categories])

            ax.invert_yaxis()

        fig.tight_layout()
        if show:
            plt.show()
        return fig

    def show(self) -> None:
        """Print information about the distribution"""
        plural = "" if len(self.dimensions) == 1 else "s"
        print(
            f'Distribution: "{self.name}" has dimension{plural} '
            f'{self.dimensions}, x_label: "{self.x_label}", '
            f'y_label: "{self.y_label}", '
            f'distribution_type: "{self.distribution_type}"'
        )

    @classmethod
    def plot_multiple(
        cls,
        dists: List["Distribution"],
        show: bool = True,
        options: Optional[Union[str, List[str]]] = None,
        colors: Optional[Union[str, List[str]]] = None,
    ) -> Optional[Figure]:
        """
        Plot multiple distributions into a single Figure using vertically
        stacked subplots.

        Returns
        -------
        Figure or None
            A matplotlib Figure when at least one distribution is plotted;
            otherwise None when `dists` is empty.

        Parameters
        ----------
        dists : list[Distribution]
            Distributions to plot. 2-D are rendered as heatmaps; 1-D below them.
        show : bool
            Whether to call ``plt.show()`` at the end.
        options : str | list[str] | None
            plot style per distribution (e.g. "bar" or "line"). If a single
            string is given, it is broadcast to all distributions. If None,
            defaults to "bar".
        colors : str | list[str] | None
            color option per distribution. If a single string is given, it is
            broadcast to all 1-D distributions. If None, defaults to
            the single color Distribution.DEFAULT_BAR_COLOR.

        Notes
        -----
        - distributions are plotted in the same order they were presented in
          dists list
        - as long as a Distribution or inherited class has a valid plot function
          implemented, the relevant plot will be added to the figure at the
          specified axes.
        - `options` and `colors` apply to all distributions
        - Although the original plot function is only limited to
          `option` and `color` being used in the 1-D case, it is not to say
          that a class inheriting Distribution won't leverage these arguments.
        - You can pass either a list (per-series) or a single string. When a
          single string is provided, it will be broadcast to all inputs.
          For example, kinds="line" makes all 1-D plots line charts.
        """
        if not dists:
            return None

        # when single string, broadcast to all distributions
        options = options or ["bar"] * len(dists)
        colors = colors or [Distribution.DEFAULT_BAR_COLOR] * len(dists)
        if isinstance(options, str):
            options = [options] * len(dists)
        if isinstance(colors, str):
            colors = [colors] * len(dists)
        if len(options) != len(dists) or len(colors) != len(dists):
            raise ValueError(
                "kinds/colors must match number of distributions in list case"
            )

        # Create a vertical stack of subplots sized to total count
        fig, axes = plt.subplots(len(dists), 1, squeeze=False)
        axes = axes.ravel()
        # use an axes iterator here
        ax_iter = iter(axes)
        for d, k, c in zip(dists, options, colors):
            ax = next(ax_iter)
            d.plot(color=c, option=k, show=False, fig=fig, ax=ax)

        fig.tight_layout()
        if show:
            plt.show()
        return fig

    @classmethod
    def plot_grouped_1d(
        cls,
        dists: List["Distribution"],
        show: bool = True,
        options: Optional[Union[str, List[str]]] = None,
        colors: Optional[Union[str, List[str]]] = None,
    ) -> Optional[Figure]:
        """Overlay multiple 1-D distributions on a single axes.

        This function draws all input 1-D distributions in one matplotlib
        Axes so that each category (x bin) shows a "group" of values—one
        per distribution. You can mix plotting styles using the `kinds`
        argument (for example, some as bars and others as lines with
        markers. Colors are controlled via the `colors` argument.

        Parameters
        ----------
        dists : list[Distribution]
            1-D distributions to compare in a single plot.
        show : bool
            Whether to call ``plt.show()`` at the end.
        options : str | list[str] | None
            Per-distribution plot style. Allowed values: "bar" or "line".
            You can provide a single string to apply to all series (broadcast),
            or a list with length `len(dists)`. If None, all series default to
            "bar".
        colors : str | list[str] | None
            Per-distribution color list. You can provide a single string to
            apply to all series (broadcast), or a list with length `len(dists)`.
            If None, a distinct default color palette is applied (rcParams cycle
            or the tab10 palette).

        Returns
        -------
        Figure or None
            A matplotlib Figure if any distributions are plotted; None when
            `dists` is empty.

        Constraints
        -----------
        - Only 1-D distributions are accepted. All inputs must have the same
          length (number of categories) so they can be grouped per category.
        - The x/y labels and category names are taken from the first
          distribution in `dists`. Hence, this function does not support
          overlaying 1-D distributions with different categories and labels.

        How this differs from plot_multiple
        -----------------------------------
        - plot_grouped_1d overlays all 1-D distributions on a single axes
          to allow:
            1. per-category (bin-by-bin) comparison intuitive and compact
               for grouped bar graphs
            2. intuitive and compact gradient comparison for overlaid line
               graphs.

          Since all distributions are plotted in a single plot, we can
          compare all plots within a single legend.
        - plot_multiple creates a vertical stack of subplots, one per
          distribution, while leveraging the plot attribute of each
          Distribution (and also supports 2-D heatmaps).
        """
        # Validate inputs
        if not dists:
            return None
        if any(len(d.dimensions) != 1 for d in dists):
            raise ValueError(
                "All distributions must be 1-D for grouped plotting"
            )
        # number of categories for each plot in the 1d distribution
        dimension = dists[0].dimensions[0]
        if any(d.dimensions[0] != dimension for d in dists):
            raise ValueError("All 1-D distributions must have the same length")
        # labels and categories will need to be the same...
        # or else some of the data visualization for axes will be misleading
        # since this function does not support plotting multiple axes labels
        # and categories on the same plot
        if any(
            d.x_label != dists[0].x_label or d.y_label != dists[0].y_label
            for d in dists
        ):
            raise ValueError("All 1-D distributions must have same axes labels")
        if any(
            d.x_categories != dists[0].x_categories
            or d.y_categories != dists[0].y_categories
            for d in dists
        ):
            raise ValueError(
                "All 1-D distributions must have same axes categories"
            )

        # when single string, broadcast to all
        if isinstance(options, str):
            options = [options] * len(dists)
        if isinstance(colors, str):
            colors = [colors] * len(dists)
        if options is None:
            options = ["bar"] * len(dists)
        if colors is None:
            # get the default ListedColormap; get_cmap does not always
            # return an object with .colors, so we have to ignore the type:
            base_colors = plt.get_cmap("tab10").colors  # type: ignore
            colors = [
                base_colors[i % len(base_colors)] for i in range(len(dists))
            ]
        if len(options) != len(dists) or len(colors) != len(dists):
            raise ValueError(
                "kinds and colors must match number of distributions"
            )

        bar_graph_info = None
        line_graph_info = None
        # partition bar graphs and line graphs to be plotted separately
        # (so that line graphs don't each take up a bin themselves)
        if isinstance(options, list):
            bar_graph_info = [
                (dist, color)
                for dist, kind, color in zip(dists, options, colors)
                if kind == "bar"
            ]
            line_graph_info = [
                (dist, color)
                for dist, kind, color in zip(dists, options, colors)
                if kind in ("line", "plot")
            ]

        fig, ax = plt.subplots()

        # Grouped bar arithmetic (unit bar width, grouped per category)
        # must have at least 1 bin for the line plot to be valid
        n = max(len(bar_graph_info), 1)
        # bar_width does not matter here, since everything in the grouped bar
        # graph is scaled according to this variable
        bar_width = 1
        x_coords = np.arange(dimension) * bar_width * n
        bottom_half, upper_half = n // 2, n - n // 2
        width_idxes = range(-bottom_half, upper_half + 1)
        is_even_offset = ((n + 1) % 2) * bar_width / 2

        # setting plot axes
        ax.set_xticks(x_coords)
        ax.set_xticklabels([str(d) for d in dists[0].x_categories])
        ax.set_xlabel(dists[0].x_label)
        ax.set_ylabel(dists[0].y_label)
        ax.set_title("Grouped Histogram Plot for 1-D Distributions")

        for width_idx, (dist, color) in zip(width_idxes, bar_graph_info):
            x_axis = x_coords + width_idx * bar_width + is_even_offset
            ax.bar(
                x_axis, dist.data, width=bar_width, label=dist.name, color=color
            )

        for dist, color in line_graph_info:
            ax.plot(
                x_coords, dist.data, color=color, marker="o", label=dist.name
            )

        ax.legend()
        fig.tight_layout()
        if show:
            plt.show()
        return fig



"""
The `pitch_list_transformations` module
provides basic functions for transforming pitch lists
(e.g., transposition, inversion, retrograde, rotation).

This small module provides some more niche routines
that are specific to tone rows in serial music.

"""

from typing import List, Tuple, Union

from amads.pitch.transformations import every_nth, transpose_to


def rotate_hexachords(
    row: Union[List, Tuple], transpose_iterations: bool = False
) -> list:
    """
    Implements a set of hexachord rotations of the kind described in krenek 1960, p.212.
    Splits the row into two hexachords and iteratively rotates each.
    This function returns a list of lists with each iteration until
    the cycle is complete and come full circle.


    Parameters
    ----------
    row
        A tone row, or any sequence of 12 integers.
    transpose_iterations
        If True, transpose each iteration to start on the original pitch of the hexachord.
        This alternative is also described by krenek.
        Note this often converts a 12-tone row into one with repeated pitches.

    Returns
    -------
    list
        A list of lists with the full cycle.

    Examples
    --------
    >>> row_krenek = [5, 7, 9, 10, 1, 3, 11, 0, 2, 4, 6, 8]
    >>> for x in rotate_hexachords(row_krenek):
    ...     print(x)
    [5, 7, 9, 10, 1, 3, 11, 0, 2, 4, 6, 8]
    [7, 9, 10, 1, 3, 5, 0, 2, 4, 6, 8, 11]
    [9, 10, 1, 3, 5, 7, 2, 4, 6, 8, 11, 0]
    [10, 1, 3, 5, 7, 9, 4, 6, 8, 11, 0, 2]
    [1, 3, 5, 7, 9, 10, 6, 8, 11, 0, 2, 4]
    [3, 5, 7, 9, 10, 1, 8, 11, 0, 2, 4, 6]
    [5, 7, 9, 10, 1, 3, 11, 0, 2, 4, 6, 8]
    """

    assert len(row) == 12

    rows = [row]  # initialise with starting

    hexachord1note1 = row[0]
    hexachord2note1 = row[6]

    for i in range(1, 6):
        first_hexachord = row[i:6] + row[0:i]
        second_hexachord = row[6 + i :] + row[6 : 6 + i]

        if transpose_iterations:
            first_hexachord = transpose_to(
                first_hexachord, start=hexachord1note1
            )
            second_hexachord = transpose_to(
                second_hexachord, start=hexachord2note1
            )

        new_row = first_hexachord + second_hexachord
        rows.append(new_row)

    rows.append(row)  # completes the cycle

    return rows


def pair_swap_krenek(row: Union[List, Tuple]) -> list:
    """
    Iteratively swaps pairs of adjacent notes in a row
    with a two-step process as described in Krenek 1960, p.213.

    Returns a list of 13 rows of which the last is the retrograde of the first.
    As such, calling this twice brings you back to the original row.

    Parameters
    ----------
    row
        A tone row, or any sequence of 12 integers.

    Returns
    -------
    list
        A list of lists with the full cycle.

    Examples
    --------
    >>> pair_swap_row = [9, 2, 3, 6, 5, 1, 7, 4, 8, 0, 10, 11]
    >>> for x in pair_swap_krenek(pair_swap_row):
    ...     print(x)
    [9, 2, 3, 6, 5, 1, 7, 4, 8, 0, 10, 11]
    [9, 3, 2, 5, 6, 7, 1, 8, 4, 10, 0, 11]
    [3, 9, 5, 2, 7, 6, 8, 1, 10, 4, 11, 0]
    [3, 5, 9, 7, 2, 8, 6, 10, 1, 11, 4, 0]
    [5, 3, 7, 9, 8, 2, 10, 6, 11, 1, 0, 4]
    [5, 7, 3, 8, 9, 10, 2, 11, 6, 0, 1, 4]
    [7, 5, 8, 3, 10, 9, 11, 2, 0, 6, 4, 1]
    [7, 8, 5, 10, 3, 11, 9, 0, 2, 4, 6, 1]
    [8, 7, 10, 5, 11, 3, 0, 9, 4, 2, 1, 6]
    [8, 10, 7, 11, 5, 0, 3, 4, 9, 1, 2, 6]
    [10, 8, 11, 7, 0, 5, 4, 3, 1, 9, 6, 2]
    [10, 11, 8, 0, 7, 4, 5, 1, 3, 6, 9, 2]
    [11, 10, 0, 8, 4, 7, 1, 5, 6, 3, 2, 9]

    """

    rows = [row]

    for pair in range(6):

        # First swap operation: starting at position 1 (2nd pitch)
        row = [x for x in row]
        for x in range(1, 11, 2):
            row[x], row[x + 1] = row[x + 1], row[x]
        rows.append(row)

        # Second swap operation: starting at position 0 (1st pitch)
        row = [x for x in row]
        for x in range(0, 12, 2):
            row[x], row[x + 1] = row[x + 1], row[x]
        rows.append(row)

    return rows


def lumsdaine_cycle(
    row: Union[list, None] = None,
) -> list[list]:
    """
    A multipart
    rotation and re-combination
    method as reported in
    Hopper's "The Music of David Lumsdaine", p.21.

    Parameters
    ----------
    row
        A tone row, or any sequence of 12 integers.

    Returns
    -------
    list
        A list of lists with the full cycle.

    Examples
    --------
    >>> for x in lumsdaine_cycle():
    ...     print(x)
    [11, 6, 5, 0, 3, 2, 9, 8, 10, 4, 1, 7]
    [11, 9, 6, 8, 5, 10, 0, 4, 3, 1, 2, 7]
    [11, 10, 2, 8, 3, 9, 0, 7, 5, 1, 6, 4]
    [11, 0, 10, 7, 2, 5, 8, 1, 3, 6, 9, 4]
    [2, 6, 10, 1, 11, 5, 9, 7, 3, 0, 8, 4]
    [2, 9, 6, 7, 10, 3, 1, 0, 11, 8, 5, 4]
    [2, 3, 5, 7, 11, 9, 1, 4, 10, 8, 6, 0]
    [2, 1, 3, 4, 5, 10, 7, 8, 11, 6, 9, 0]
    [5, 6, 3, 8, 2, 10, 9, 4, 11, 1, 7, 0]
    [5, 9, 6, 4, 3, 11, 8, 1, 2, 7, 10, 0]
    [5, 11, 10, 4, 2, 9, 8, 0, 3, 7, 6, 1]
    [5, 8, 11, 0, 10, 3, 4, 7, 2, 6, 9, 1]
    [10, 6, 11, 7, 5, 3, 9, 0, 2, 8, 4, 1]
    [10, 9, 6, 0, 11, 2, 7, 8, 5, 4, 3, 1]
    [10, 2, 3, 0, 5, 9, 7, 1, 11, 4, 6, 8]
    [10, 7, 2, 1, 3, 11, 0, 4, 5, 6, 9, 8]
    [3, 6, 2, 4, 10, 11, 9, 1, 5, 7, 0, 8]
    [3, 9, 6, 1, 2, 5, 4, 7, 10, 0, 11, 8]
    [3, 5, 11, 1, 10, 9, 4, 8, 2, 0, 6, 7]
    [3, 4, 5, 8, 11, 2, 1, 0, 10, 6, 9, 7]

    """
    if row is None:
        row = [11, 6, 5, 0, 3, 2, 9, 8, 10, 4, 1, 7]
    out = []
    for i in range(5):  # run 4, update to new starting row, and run again
        out += lumsdaine_4(row)
        row = every_nth(out[-1], start_index=4)
    return out


def lumsdaine_4(
    row: Union[List, None] = None,
) -> list[list]:
    """
    One phase of `lumsdaine_cycle()`
    producing a set of 4 rows.

    Parameters
    ----------
    row
        A tone row, or any sequence of 12 integers.

    Returns
    -------
    list
        A list of lists with the full cycle.

    Examples
    --------
    >>> for x in lumsdaine_4([11, 6, 5, 0, 3, 2, 9, 8, 10, 4, 1, 7]):
    ...     print(x)
    [11, 6, 5, 0, 3, 2, 9, 8, 10, 4, 1, 7]
    [11, 9, 6, 8, 5, 10, 0, 4, 3, 1, 2, 7]
    [11, 10, 2, 8, 3, 9, 0, 7, 5, 1, 6, 4]
    [11, 0, 10, 7, 2, 5, 8, 1, 3, 6, 9, 4]

    >>> for x in lumsdaine_4([5, 9, 7, 3, 0, 8, 4, 2, 6, 10, 1, 11]):
    ...     print(x)
    [5, 9, 7, 3, 0, 8, 4, 2, 6, 10, 1, 11]
    [5, 4, 9, 2, 7, 6, 3, 10, 0, 1, 8, 11]
    [5, 6, 8, 2, 0, 4, 3, 11, 7, 1, 9, 10]
    [5, 3, 6, 11, 8, 7, 2, 1, 0, 9, 4, 10]

    """
    if row is None:
        row = [11, 6, 5, 0, 3, 2, 9, 8, 10, 4, 1, 7]
    row2 = lumsdaine_hexachord_pairs(row)
    row3 = every_nth(row2)
    row4 = lumsdaine_hexachord_pairs(row3)
    return [row, row2, row3, row4]


def lumsdaine_hexachord_pairs(row: list):
    """
    Constituent step in the Lumsdaine method
    that re-arrange row elements indices into the order 0, 6, 1, 7 ...

    Parameters
    ----------
    row
        A tone row, or any sequence of 12 integers.

    Returns
    -------
    list
        A list of lists with the full cycle.

    Examples
    --------
    >>> lumsdaine_hexachord_pairs([11, 6, 5, 0, 3, 2, 9, 8, 10, 4, 1, 7])
    [11, 9, 6, 8, 5, 10, 0, 4, 3, 1, 2, 7]
    """
    out_row = []
    for i in range(6):
        out_row += [row[i], row[i + 6]]
    return out_row


# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()
